from __future__ import annotations
import numpy
import typing

__all__ = [
    "AltitudeGrid",
    "AtmosphereStokes_1",
    "AtmosphereStokes_3",
    "AtmosphereStorageStokes_1",
    "AtmosphereStorageStokes_3",
    "BRDFStokes_1",
    "BRDFStokes_3",
    "Config",
    "ConstantSpacing",
    "Coordinates",
    "DerivativeMapping",
    "Disabled",
    "DiscreteOrdinates",
    "Ellipsoidal",
    "EmissionSource",
    "EngineStokes_1",
    "EngineStokes_3",
    "Exact",
    "Geodetic",
    "Geometry1D",
    "GeometryType",
    "GridSpacing",
    "GroundViewingSolar",
    "InputValidationMode",
    "InterpolationMethod",
    "LambertianStokes_1",
    "LambertianStokes_3",
    "LinearInterpolation",
    "LinearSpacing",
    "LinearizedMie",
    "LowerInterpolation",
    "MODISStokes_1",
    "MODISStokes_3",
    "MieData",
    "MieOutput",
    "MultipleScatterSource",
    "NoSource",
    "Observer",
    "OccultationSource",
    "OutOfBoundsExtend",
    "OutOfBoundsPolicy",
    "OutOfBoundsSetZero",
    "OutputDerivMappedStokes_1",
    "OutputDerivMappedStokes_3",
    "OutputIdealStokes_1",
    "OutputIdealStokes_3",
    "OutputStokes_1",
    "OutputStokes_3",
    "PlaneParallel",
    "PseudoSpherical",
    "ShellInterpolation",
    "SingleScatterSource",
    "SnowKokhanovskyStokes_1",
    "SnowKokhanovskyStokes_3",
    "Solar",
    "Source",
    "Spherical",
    "Standard",
    "StokesBasis",
    "Strict",
    "SuccessiveOrders",
    "SurfaceDerivativeMapping",
    "SurfaceStokes_1",
    "SurfaceStokes_3",
    "Table",
    "TangentAltitudeSolar",
    "ThreadingModel",
    "TwoStream",
    "ViewingGeometry",
    "ViewingGeometryBase",
    "Wavelength",
    "WignerD",
    "voigt_broaden",
]

class AltitudeGrid:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        arg0: numpy.ndarray[numpy.float64[m, 1]],
        arg1: GridSpacing,
        arg2: OutOfBoundsPolicy,
        arg3: InterpolationMethod,
    ) -> None: ...

class AtmosphereStokes_1:
    storage: ...
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self, arg0: int, arg1: Geometry1D, arg2: Config, arg3: bool
    ) -> None: ...
    def apply_delta_m_scaling(self, arg0: int) -> None: ...
    @property
    def surface(self) -> ...: ...

class AtmosphereStokes_3:
    storage: ...
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self, arg0: int, arg1: Geometry1D, arg2: Config, arg3: bool
    ) -> None: ...
    def apply_delta_m_scaling(self, arg0: int) -> None: ...
    @property
    def surface(self) -> ...: ...

class AtmosphereStorageStokes_1:
    d_leg_coeff: numpy.ndarray[numpy.float64[..., ..., ..., ...]]
    emission_source: numpy.ndarray[numpy.float64[m, n]]
    leg_coeff: numpy.ndarray[numpy.float64[..., ..., ...]]
    solar_irradiance: numpy.ndarray[numpy.float64[m, 1]]
    ssa: numpy.ndarray[numpy.float64[m, n]]
    total_extinction: numpy.ndarray[numpy.float64[m, n]]
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...
    def get_derivative_mapping(self, name: str) -> DerivativeMapping: ...
    def resize_derivatives(self, num_deriv: int) -> None: ...
    @property
    def d_f(self) -> numpy.ndarray[numpy.float64[..., ..., ...]]: ...
    @property
    def derivative_mappings(self) -> dict[str, DerivativeMapping]: ...
    @property
    def f(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class AtmosphereStorageStokes_3:
    d_leg_coeff: numpy.ndarray[numpy.float64[..., ..., ..., ...]]
    emission_source: numpy.ndarray[numpy.float64[m, n]]
    leg_coeff: numpy.ndarray[numpy.float64[..., ..., ...]]
    solar_irradiance: numpy.ndarray[numpy.float64[m, 1]]
    ssa: numpy.ndarray[numpy.float64[m, n]]
    total_extinction: numpy.ndarray[numpy.float64[m, n]]
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...
    def get_derivative_mapping(self, name: str) -> DerivativeMapping: ...
    def resize_derivatives(self, num_deriv: int) -> None: ...
    @property
    def d_f(self) -> numpy.ndarray[numpy.float64[..., ..., ...]]: ...
    @property
    def derivative_mappings(self) -> dict[str, DerivativeMapping]: ...
    @property
    def f(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class BRDFStokes_1:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...

class BRDFStokes_3:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...

class Config:
    stokes_basis: StokesBasis
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None:
        """
        Object which stores all of the configuration settings for the radiative transfer calculation.
        """

    @property
    def delta_m_scaling(self) -> bool:
        """
        Controls whether the delta-M scaling is applied to the calculation.  Defaults to False.
        """

    @delta_m_scaling.setter
    def delta_m_scaling(self, arg1: bool) -> None: ...
    @property
    def do_backprop(self) -> bool:
        """
        Enables backpropagation for the weighting functions when using the DO source in plane parallel or pseudo-spherical geometry.
        Can greatly improve the computation speed of the calculation when the number of lines of sight is small. Default to True
        """

    @do_backprop.setter
    def do_backprop(self, arg1: bool) -> None: ...
    @property
    def emission_source(self) -> EmissionSource:
        """
        Sets which (if any) emission source is to be used inside the calculation.

        `sasktran2.EmissionSource.NoSource` (Default)
            No emission source included

        `sasktran2.EmissionSource.Standard`
            An emission source defined on the atmosphere grid is enabled.
        """

    @emission_source.setter
    def emission_source(self, arg1: EmissionSource) -> None: ...
    @property
    def init_successive_orders_with_discrete_ordinates(self) -> bool:
        """
        If set to true, when using the successive orders source, it will be initialized with a source calculated with
        the discrete ordinates source instead of a single scattering source.  This greatly reduces the number
        of iterations required for the method, as well as provides a better weighting function approximation.
        """

    @init_successive_orders_with_discrete_ordinates.setter
    def init_successive_orders_with_discrete_ordinates(self, arg1: bool) -> None: ...
    @property
    def input_validation_mode(self) -> InputValidationMode:
        """
        Sets the input validation mode to use in the calculation.

        `sasktran2.InputValidationMode.Strict` (Default)
            All input validation checks are performed. This is the recommended mode for most users.

        `sasktran2.InputValidationMode.Standard`
            Only the most important input validation checks are performed. This mode is recommended
            for advanced users who are confident in the input data.

        `sasktran2.InputValidationMode.Disabled`
            No input validation checks are performed. This mode is recommended for advanced users who
            are confident in the input data and want to maximize performance.
        """

    @input_validation_mode.setter
    def input_validation_mode(self, arg1: InputValidationMode) -> None: ...
    @property
    def los_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the observer line of sight rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Defaults to False.
        """

    @los_refraction.setter
    def los_refraction(self, arg1: bool) -> None: ...
    @property
    def multiple_scatter_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the multiple scatter source. Requires
        the refractive index to be set in the Geometry object for refraction to work.
        Only has an effect when the SuccessiveOrders multiple scatter source term is being used.
        Defaults to False.
        """

    @multiple_scatter_refraction.setter
    def multiple_scatter_refraction(self, arg1: bool) -> None: ...
    @property
    def multiple_scatter_source(self) -> MultipleScatterSource:
        """
        Sets which (if any) multiple scatter source is to be used inside the calculation.

        `sasktran2.MultipleScatterSource.NoSource` (Default)
            Multiple scattering is disabled

        `sasktran2.MultipleScatterSource.DiscreteOrdinates`
            The discrete ordinates technique is used to estimate the multiple scatter signal

        `sasktran2.MultipleScatterSource.SuccessiveOrders`
            The successive orders of scattering method is used to estimate the multiple scatter
            signal
        """

    @multiple_scatter_source.setter
    def multiple_scatter_source(self, arg1: MultipleScatterSource) -> None: ...
    @property
    def num_forced_azimuth(self) -> int:
        """
        If set to a value greater than 0, the discrete ordinates method will use this number of azimuth terms independent of convergence.
        Defaults to -1, which means to use the number of azimuth terms required for convergence.
        """

    @num_forced_azimuth.setter
    def num_forced_azimuth(self, arg1: int) -> None: ...
    @property
    def num_singlescatter_moments(self) -> int:
        """
        The number of Legendre expansion moments to use in the single scatter calculation.
        Must be greater or equal to num_streams. Default to 16.
        """

    @num_singlescatter_moments.setter
    def num_singlescatter_moments(self, arg1: int) -> None: ...
    @property
    def num_stokes(self) -> int:
        """
        Sets the number of Stokes parameters used in the calculation. 1 is equivalent to the scalar approximation.
        Currently the only supported values are 1, and 3.
        """

    @num_stokes.setter
    def num_stokes(self, arg1: int) -> None: ...
    @property
    def num_streams(self) -> int:
        """
        The number of streams to use in the discrete ordinates method. This is the number of streams
        in full space, i.e. each hemisphere has num_streams / 2 angular discretizations.  Must
        be an even number. Default to 16.
        """

    @num_streams.setter
    def num_streams(self, arg1: int) -> None: ...
    @property
    def num_successive_orders_incoming(self) -> int:
        """
        The number of integration nodes to use in the successive orders algorithm when calculating the incoming
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """

    @num_successive_orders_incoming.setter
    def num_successive_orders_incoming(self, arg1: int) -> None: ...
    @property
    def num_successive_orders_iterations(self) -> int:
        """
        The number of iterations to perform when using the successive orders of scattering multiple scatter
        source.
        """

    @num_successive_orders_iterations.setter
    def num_successive_orders_iterations(self, arg1: int) -> None: ...
    @property
    def num_successive_orders_outgoing(self) -> int:
        """
        The number of sample points to use in the successive orders algorithm to calculate the outgoing source function on
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """

    @num_successive_orders_outgoing.setter
    def num_successive_orders_outgoing(self, arg1: int) -> None: ...
    @property
    def num_successive_orders_points(self) -> int:
        """
        The number of incoming points to use in the successive orders calculation for each solar
        zenith angle.  Must be equal to or less than the number of atmosphere altitude grid points.
        Default is -1 which means to use every altitude grid point.
        """

    @num_successive_orders_points.setter
    def num_successive_orders_points(self, arg1: int) -> None: ...
    @property
    def num_sza(self) -> int:
        """
        The number of solar zenith angle discretizations to use when calculating the multiple scatter source.
        For the discrete ordinates source, this determines the number of independent discrete ordinates calculations to perform.
        In the successive orders of scattering source, this is directly the number of discretizations.
        Defaults to 1, indicating that the multiple scatter source is estimated only at the reference point.
        """

    @num_sza.setter
    def num_sza(self, arg1: int) -> None: ...
    @property
    def num_threads(self) -> int:
        """
        Controls the number of threads used in the calculation.  For maximum performance it is
        recommended to set this to the number of physical cores on your machine.  Defaults to
        1
        """

    @num_threads.setter
    def num_threads(self, arg1: int) -> None: ...
    @property
    def occultation_source(self) -> OccultationSource:
        """
        Sets which (if any) occultation source is to be used inside the calculation.

        `sasktran2.OccultationSource.NoSource` (Default)
            No occultation source included

        `sasktran2.OccultationSource.Standard`
            A constant source of 1 is placed at the end of every individual line of sight.
        """

    @occultation_source.setter
    def occultation_source(self, arg1: OccultationSource) -> None: ...
    @property
    def single_scatter_source(self) -> SingleScatterSource:
        """
        Sets which (if any) single scatter source is to be used inside the calculation.

        `sasktran2.SingleScatterSource.Exact` (Default)
            A single scatter source where exact ray tracing is performed at each quadrature
            point along the observer lines of sight towards the sun

        `sasktran2.SingleScatterSource.Table`
            A single scatter source where a pre-computed table is used to calculate solar
            transmission to quadrature points along the line of sight.

        `sasktran2.SingleScatterSource.DiscreteOrdinates`
            Lets the discrete ordinates source function calculate the single scatter source. Only
            has an effect if the geometry mode is set to PlaneParallel or PseudoSpherical, and
            if the DiscreteOrdinates source function is also used for multiple scatter.

        `sasktran2.SingleScatterSource.NoSource`
            Disables the single scatter source
        """

    @single_scatter_source.setter
    def single_scatter_source(self, arg1: SingleScatterSource) -> None: ...
    @property
    def solar_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the solar line of sight rays. Requires
          the refractive index to be set in the Geometry object for refraction to work.  Only has an effect
          when the single scatter source term is set to Table.  Defaults to False.
        """

    @solar_refraction.setter
    def solar_refraction(self, arg1: bool) -> None: ...
    @property
    def threading_model(self) -> ThreadingModel:
        """
        Sets the multi-threading mode to use in the calculation.

        `sasktran2.ThreadingModel.Wavelength` (Default)
            Calculation is multi-threaded over the wavelength (batch) dimension only. This method
            works very well when this dimension is large, but may increase memory usage. It also
            is not very effective for a small number of wavelengths.

        `sasktran2.ThreadingModel.Source`
            Calculation is multi-threaded individually by each source function for each wavelength.
            This method is recommended when memory is a concern, or when the number of wavelengths
            is small.
        """

    @threading_model.setter
    def threading_model(self, arg1: ThreadingModel) -> None: ...

class Coordinates:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self, arg0: float, arg1: float, arg2: float, arg3: GeometryType, arg4: bool
    ) -> None: ...

class DerivativeMapping:
    assign_name: str
    interp_dim: str
    log_radiance_space: bool
    scat_deriv_index: int
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...
    def set_zero(self) -> None: ...
    @property
    def d_emission(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def d_extinction(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def d_leg_coeff(self) -> numpy.ndarray[numpy.float64[..., ..., ...]]: ...
    @property
    def d_ssa(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def interpolator(self) -> numpy.ndarray[numpy.float64[m, n]] | None: ...
    @interpolator.setter
    def interpolator(self, arg1: numpy.ndarray[numpy.float64[m, n]]) -> None: ...
    @property
    def is_scattering_derivative(self) -> bool: ...
    @property
    def num_output(self) -> int: ...
    @property
    def scat_factor(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class EmissionSource:
    """
    Members:

      NoSource

      Standard
    """

    NoSource: typing.ClassVar[EmissionSource]  # value = <EmissionSource.NoSource: 1>
    Standard: typing.ClassVar[EmissionSource]  # value = <EmissionSource.Standard: 0>
    __members__: typing.ClassVar[
        dict[str, EmissionSource]
    ]  # value = {'NoSource': <EmissionSource.NoSource: 1>, 'Standard': <EmissionSource.Standard: 0>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class EngineStokes_1:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        config: Config,
        model_geometry: Geometry1D,
        viewing_geometry: ViewingGeometry,
    ) -> None:
        """
        Internal SASKTRAN2 object which handles the radiative transfer calculation.

        Parameters
        ----------
        config: sasktran2.Config
            Configuration settings
        model_geometry: sasktran2.Geometry1D
            The model geometry
        viewing_geometry: sasktan2.ViewingGeometry
            The viewing geometry
        """

    def calculate_radiance(
        self, atmosphere: AtmosphereStokes_1, output: OutputStokes_1
    ) -> None:
        """
        Performs the radiative transfer calculation for the given atmosphere, placing the result in output

        Parameters
        ----------
        atmosphere: sasktran2.Atmosphere
            The atmosphere object

        output: sasktran2.Output
            The result to place the output inside
        """

class EngineStokes_3:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        config: Config,
        model_geometry: Geometry1D,
        viewing_geometry: ViewingGeometry,
    ) -> None:
        """
        Internal SASKTRAN2 object which handles the radiative transfer calculation.

        Parameters
        ----------
        config: sasktran2.Config
            Configuration settings
        model_geometry: sasktran2.Geometry1D
            The model geometry
        viewing_geometry: sasktan2.ViewingGeometry
            The viewing geometry
        """

    def calculate_radiance(
        self, atmosphere: AtmosphereStokes_3, output: OutputStokes_3
    ) -> None:
        """
        Performs the radiative transfer calculation for the given atmosphere, placing the result in output

        Parameters
        ----------
        atmosphere: sasktran2.Atmosphere
            The atmosphere object

        output: sasktran2.Output
            The result to place the output inside
        """

class Geodetic:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, equatorial_radius: float, flattening_factor: float) -> None:
        """
        A geodetic object based on a given equatorial (semi-major) radius and flattening factor.

        Standard usage is to create a geodetic object, and then initialize it through one of the
        `from_*` methods.

        Parameters
        ----------
        equatorial_radius: float
            Radius at the equator (semi-major) of the ellipsoid.
        flattening_factor: float
            Flattening factor of the ellipsoid. This is defined as (a-b)/a, where a is the semi-major axis
            and b is the semi-minor radius
        """

    def altitude_intercepts(
        self,
        altitude: float,
        observer: numpy.ndarray[numpy.float64[3, 1]],
        look_vector: numpy.ndarray[numpy.float64[3, 1]],
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], numpy.ndarray[numpy.float64[3, 1]]]:
        """
                        Calculate the two intersections of a line of sight and an altitude.

                        Parameters
                        ----------
                        altitude : float
                            Altitude in meters.
                        observer : np.ndarray
                            Three element array containing the obsever position in geocentric coordinates.
                        look_vector : np.ndarray
                            Three element array containing a normalized look vector.

                        Returns
                        -------
                        np.ndarray
                            Three element array containing the first (entering) intercept in geocentric coordinates.
                        np.ndarray
                            Three element array containing the second (exiting) intercept in geocentric coordinates.

                        Examples
                        --------
                        >>> import sasktran2 as sk
                        >>> import numpy as np
                        >>> geodetic = sk.WGS84()
                        >>> look = geodetic.from_tangent_altitude(15322, [3.676013154788849600e+005, 1.009976313640051500e+006, \
                                                                    -6.871601202127538600e+006], [0, 0, 1])
                        >>> obs = geodetic.location
                        >>> intercept1, intercept2 = geodetic.altitude_intercepts(16000, obs, look)
                        >>> print(np.array_str(intercept1, precision=3))
                        [ 1147302.059  3152186.5   -5425360.027]
                        >>> print(np.array_str(intercept2, precision=3))
                        [ 1201098.489  3299990.978 -5325574.803]
        """

    def from_lat_lon_alt(
        self, latitude: float, longitude: float, altitude: float
    ) -> None:
        """
        Initializes the Geodetic based on a specifiec latitude, longitude, and altitude.

        Parameters
        ----------
        latitude : float
            Latitude in degrees (-90 to 90)
        longitude : float
            Longitude in degrees (0 to 360 or -180 to 180)
        altitude : float
            Altitude above the geoid in metres

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> geodetic.from_lat_lon_alt(latitude=-15, longitude=-20, altitude=7342)
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -15.0, Longitude: 340.0, Altitude: 7342.0
        """

    def from_tangent_altitude(
        self,
        altitude: float,
        observer: numpy.ndarray[numpy.float64[3, 1]],
        boresight: numpy.ndarray[numpy.float64[3, 1]],
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
                        Initialized the Geodetic from a specified tangent altitude, obsever location, and bore sight plane.

                        Parameters
                        ----------
                        altitude : float
                            Tangent altitude in meters
                        observer : np.ndarray
                            Three element array containing the obsever position in geocentric coordinates
                        boresight : np.ndarray
                            Three element array containing a normalized look vector that is within the bore sight plane.

                        Returns
                        -------
                        np.ndarray
                            Three element array containing the normalized look vector to the tangent point.

                        Examples
                        --------
                        >>> import sasktran2 as sk
                        >>> geodetic = sk.WGS84()
                        >>> look = geodetic.from_tangent_altitude(15322, [ 3.676013154788849600e+005, 1.009976313640051500e+006,\
                                                                          -6.871601202127538600e+006], [0, 0, 1])
                        >>> print(look)
                        [0.28880556 0.79348676 0.53569591]
                        >>> print(geodetic)
                        WGS84 Location:
                        Latitude: -57.60888188776806, Longitude: 70.00000000000001, Altitude: 15321.971935882739
        """

    def from_tangent_point(
        self,
        observer: numpy.ndarray[numpy.float64[3, 1]],
        look_vector: numpy.ndarray[numpy.float64[3, 1]],
    ) -> None:
        """
                        Initializes  the Geodetic by calculating the tangent point from an observer position and look vector

                        Parameters
                        ----------
                        observer : np.ndarray
                            Three element array containing the observer position in geocentric coordinates
                        look_vector : np.ndarray
                            Three element array containing a normalized look vector

                        Examples
                        --------
                        >>> import sasktran2 as sk
                        >>> geodetic = sk.WGS84()
                        >>> geodetic.from_tangent_point([ 3.676013154788849600e+005, 1.009976313640051500e+006,\
                                                         -6.871601202127538600e+006], [ 2.884568631765662100e-001,\
                                                          7.925287180643269000e-001,  5.372996083468238900e-001])
                        >>> print(geodetic)
                        WGS84 Location:
                        Latitude: -57.500000192733594, Longitude: 70.0, Altitude: 10002.99586173162
        """

    def from_xyz(self, location: numpy.ndarray[numpy.float64[3, 1]]) -> None:
        """
        Initializes the Geodetic from a geocentric location

        Parameters
        ----------
        location : np.ndarray
            Three element vector containing a location in geocentric coordinates

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> geodetic.from_xyz([ 5797230.47518212, -2110019.3341472, -1642001.16317228])
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -14.999999973747736, Longitude: 340.00000000000006, Altitude: 7344.999610390202
        """

    @property
    def altitude(self) -> float:
        """
        Altitude in [m] above the surface of the ellipsoid.
        """

    @property
    def latitude(self) -> float:
        """
        Geodetic latitude in degrees
        """

    @property
    def local_south(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        A unit vector pointing in the local south direction
        """

    @property
    def local_up(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        A unit vector pointing up (perpindicular to the ellipsoidal surface)
        """

    @property
    def local_west(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        A unit vector pointing in the local west direction
        """

    @property
    def location(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Geocentric location in cartesian coordinates
        """

    @property
    def longitude(self) -> float:
        """
        Geodetic longitude in degrees
        """

    @property
    def valid(self) -> bool:
        """
        True if the geodetic object has been initialized, False otherwise.
        """

class Geometry1D:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        cos_sza: float,
        solar_azimuth: float,
        earth_radius_m: float,
        altitude_grid_m: numpy.ndarray[numpy.float64[m, 1]],
        interpolation_method: InterpolationMethod,
        geometry_type: GeometryType,
    ) -> None:
        """
        Initializes a geometry where the atmosphere varies only in 1 dimension (altitude).  The reference point
        is defined by solar angles at the reference point.

        Parameters
        ----------
        cos_sza: float
            Cosine of solar zenith angle at the reference point
        solar_azimuth: float
            Solar azimuth angle at the reference point.
        earth_radius_m: float
            Radius of the earth.  Only has an effect if geometry_type is not set to PlaneParallel
        altitude_grid_m: np.array
            One dimensional altitude grid
        interpolation_method: sasktran2.InterpolationMethod
            The interpolation method to use in-between geometry grid points

            `sasktran2.InterpolationMethod.LinearInterpolation`
                In-between grid points, linear interpolation is assumed.  This means that Atmospheric quantities
                such as extinction, single scatter albedo, should be thought of as sampled on the geometry grid points.

            `sasktran2.InterpolationMethod.ShellInterpolation`
                Atmospheric quantities such as extinction, single scatter albedo, are assumed to be constant in-between
                geometry grid points.

        geometry_type: sasktran2.GeometryType
            The global geometry type used inside the radiative transfer calculation.

            `sasktran2.GeometryType.Spherical`
                All aspects of the calculation are done using spherical geometry.

            `sasktran2.GeometryType.PlaneParallel`
                All aspects of the calculation are done using plane-parallel geometry.

            `sasktran2.GeometryType.PseudoSpherical`
                Line of sight integration and the multiple scatter calculation is done using
                plane parallel geometry, however the initial solar source function is calculated
                using a spherical geometry.
        """

    def altitudes(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @property
    def refractive_index(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        The refractive index of the atmosphere.  This is used to calculate refraction in the radiative transfer calculation.
        Defaults to 1.0 which indicates no refractive effects.  Only has an effect if the refraction configuration options are
        enabled in the `sasktran2.Config` object.
        """

    @refractive_index.setter
    def refractive_index(self, arg1: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...

class GeometryType:
    """
    Members:

      PlaneParallel

      Spherical

      PseudoSpherical

      Ellipsoidal
    """

    Ellipsoidal: typing.ClassVar[GeometryType]  # value = <GeometryType.Ellipsoidal: 3>
    PlaneParallel: typing.ClassVar[
        GeometryType
    ]  # value = <GeometryType.PlaneParallel: 0>
    PseudoSpherical: typing.ClassVar[
        GeometryType
    ]  # value = <GeometryType.PseudoSpherical: 1>
    Spherical: typing.ClassVar[GeometryType]  # value = <GeometryType.Spherical: 2>
    __members__: typing.ClassVar[
        dict[str, GeometryType]
    ]  # value = {'PlaneParallel': <GeometryType.PlaneParallel: 0>, 'Spherical': <GeometryType.Spherical: 2>, 'PseudoSpherical': <GeometryType.PseudoSpherical: 1>, 'Ellipsoidal': <GeometryType.Ellipsoidal: 3>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class GridSpacing:
    """
    Members:

      ConstantSpacing

      LinearSpacing
    """

    ConstantSpacing: typing.ClassVar[
        GridSpacing
    ]  # value = <GridSpacing.ConstantSpacing: 0>
    LinearSpacing: typing.ClassVar[
        GridSpacing
    ]  # value = <GridSpacing.LinearSpacing: 1>
    __members__: typing.ClassVar[
        dict[str, GridSpacing]
    ]  # value = {'ConstantSpacing': <GridSpacing.ConstantSpacing: 0>, 'LinearSpacing': <GridSpacing.LinearSpacing: 1>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class GroundViewingSolar(ViewingGeometryBase):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        cos_sza: float,
        relative_azimuth: float,
        cos_viewing_zenith: float,
        observer_altitude_m: float,
    ) -> None:
        """
        Defines a viewing ray that is looking at the ground from angles defined at the ground location. Note that
        all of these parameters assumes straight line paths (i.e. no atmospheric refraction)

        Parameters
        ----------
        cos_sza: float
            Cosine of solar zenith angle at the ground point [unitless]
        relative_azimuth: float
            Relative azimuth angle to the sun [rad] at the ground point. An angle of 0 degrees corresponds to the forward scattering plane.
        observer_altitude_m: float
            Observer altitude relative to the earth [m]
        cos_viewing_zenith: float
            Cosine of the viewing zenith angle at the ground point [unitless]
        """

class InputValidationMode:
    """
    Members:

      Strict

      Standard

      Disabled
    """

    Disabled: typing.ClassVar[
        InputValidationMode
    ]  # value = <InputValidationMode.Disabled: 2>
    Standard: typing.ClassVar[
        InputValidationMode
    ]  # value = <InputValidationMode.Standard: 1>
    Strict: typing.ClassVar[
        InputValidationMode
    ]  # value = <InputValidationMode.Strict: 0>
    __members__: typing.ClassVar[
        dict[str, InputValidationMode]
    ]  # value = {'Strict': <InputValidationMode.Strict: 0>, 'Standard': <InputValidationMode.Standard: 1>, 'Disabled': <InputValidationMode.Disabled: 2>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class InterpolationMethod:
    """
    Members:

      ShellInterpolation

      LinearInterpolation

      LowerInterpolation
    """

    LinearInterpolation: typing.ClassVar[
        InterpolationMethod
    ]  # value = <InterpolationMethod.LinearInterpolation: 1>
    LowerInterpolation: typing.ClassVar[
        InterpolationMethod
    ]  # value = <InterpolationMethod.LowerInterpolation: 2>
    ShellInterpolation: typing.ClassVar[
        InterpolationMethod
    ]  # value = <InterpolationMethod.ShellInterpolation: 0>
    __members__: typing.ClassVar[
        dict[str, InterpolationMethod]
    ]  # value = {'ShellInterpolation': <InterpolationMethod.ShellInterpolation: 0>, 'LinearInterpolation': <InterpolationMethod.LinearInterpolation: 1>, 'LowerInterpolation': <InterpolationMethod.LowerInterpolation: 2>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class LambertianStokes_1(BRDFStokes_1):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class LambertianStokes_3(BRDFStokes_3):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class LinearizedMie:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, num_threads: int = 1) -> None:
        """
        A Mie object created with no input parameters.

        Standard usage is to create a Mie object, and then calculate mie parameters using
        `calculate` method.

        Parameters
        ----------
        num_threads : int
            Number of threads to use for the Mie calculation. Default is 1.
        """

    def calculate(
        self,
        size_param: numpy.ndarray[numpy.float64[m, 1]],
        refractive_index: complex,
        cos_angles: numpy.ndarray[numpy.float64[m, 1]],
        calculate_derivative: bool,
    ) -> MieOutput:
        """
        Performs the Mie computation for an array of size parameters, a single refractive index, and an array that is the cosine of the scattering angles.

        Parameters
        ----------
        size_param : np.ndarray
            Array of Mie size parameters. Shape (size).
        refractive_index : complex
            Complex Mie refractive index
        cos_angles : np.ndarray
            Array of cosine of angles to calculate the scattering amplitude at. Shape (angle).
        calculate_derivative : bool, optional
            Optional parameter, initiates calculations of derivatives for size parameter and refractive index (not implemented at the moment), by default False

        Returns
        -------
        MieOutput
            MieOutput that contains the original size parameters, cosine of angles, and refractive index, as well as the calculated mie parameters.

        Examples
        --------

        >>> import sasktran2 as sk
        >>> import numpy as np
        >>> mie = sk.mie.LinearizedMie()
        >>> size_param = np.array([3.0, 4.0, 5.0])
        >>> cos_angles = np.linspace(-1, 1, 100)
        >>> refractive_index = 1.5 + 0.0j
        >>> output = mie.calculate(size_param, refractive_index, cos_angles, True)

        >>> print(output.values.Qext)
        [3.41805617 4.05245221 3.92782673]
        >>> print(output.values.Qsca)
        [3.41805617 4.05245221 3.92782673]
        """

class MODISStokes_1(BRDFStokes_1):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class MODISStokes_3(BRDFStokes_3):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class MieData:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    @property
    def Qext(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Calculated Extinction Efficiency factor [unitless] for given size parameters and refractive index. Shape (size).
        """

    @property
    def Qsca(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Calculated Scattering Efficiency factor [unitless] for given size parameters and refractive index. Shape (size).
        """

    @property
    def S1(self) -> numpy.ndarray[numpy.complex128[m, n]]:
        """
        Calculated Complex Scattering Amplitude [unitless] in first direction of incident polarization for given size parameters, cos(scattering angles) and refractive index. Shape (size, angle).
        """

    @property
    def S2(self) -> numpy.ndarray[numpy.complex128[m, n]]:
        """
        Calculated Complex Scattering Amplitude [unitless] in second direction of incident polarization for given size parameters, cos(scattering angles) and refractive index. Shape (size, angle).
        """

class MieOutput:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    @property
    def cos_angles(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Array containing the cosine of the scattering angles. Shape (angle).
        """

    @property
    def refractive_index(self) -> complex:
        """
        Complex refractive index of spheres.
        """

    @property
    def size_parameter(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Array containing size parameters of spheres (2pi*radius/wavelength). Shape (size).
        """

    @property
    def values(self) -> MieData:
        """
        MieData structure containing Extinction Efficiency, Scattering Efficiency and Scattering Amplitudes.
        """

class MultipleScatterSource:
    """
    Members:

      DiscreteOrdinates

      TwoStream

      SuccessiveOrders

      NoSource
    """

    DiscreteOrdinates: typing.ClassVar[
        MultipleScatterSource
    ]  # value = <MultipleScatterSource.DiscreteOrdinates: 0>
    NoSource: typing.ClassVar[
        MultipleScatterSource
    ]  # value = <MultipleScatterSource.NoSource: 3>
    SuccessiveOrders: typing.ClassVar[
        MultipleScatterSource
    ]  # value = <MultipleScatterSource.SuccessiveOrders: 1>
    TwoStream: typing.ClassVar[
        MultipleScatterSource
    ]  # value = <MultipleScatterSource.TwoStream: 2>
    __members__: typing.ClassVar[
        dict[str, MultipleScatterSource]
    ]  # value = {'DiscreteOrdinates': <MultipleScatterSource.DiscreteOrdinates: 0>, 'TwoStream': <MultipleScatterSource.TwoStream: 2>, 'SuccessiveOrders': <MultipleScatterSource.SuccessiveOrders: 1>, 'NoSource': <MultipleScatterSource.NoSource: 3>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class OccultationSource:
    """
    Members:

      NoSource

      Standard
    """

    NoSource: typing.ClassVar[
        OccultationSource
    ]  # value = <OccultationSource.NoSource: 1>
    Standard: typing.ClassVar[
        OccultationSource
    ]  # value = <OccultationSource.Standard: 0>
    __members__: typing.ClassVar[
        dict[str, OccultationSource]
    ]  # value = {'NoSource': <OccultationSource.NoSource: 1>, 'Standard': <OccultationSource.Standard: 0>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class OutOfBoundsPolicy:
    """
    Members:

      OutOfBoundsExtend

      OutOfBoundsSetZero
    """

    OutOfBoundsExtend: typing.ClassVar[
        OutOfBoundsPolicy
    ]  # value = <OutOfBoundsPolicy.OutOfBoundsExtend: 0>
    OutOfBoundsSetZero: typing.ClassVar[
        OutOfBoundsPolicy
    ]  # value = <OutOfBoundsPolicy.OutOfBoundsSetZero: 1>
    __members__: typing.ClassVar[
        dict[str, OutOfBoundsPolicy]
    ]  # value = {'OutOfBoundsExtend': <OutOfBoundsPolicy.OutOfBoundsExtend: 0>, 'OutOfBoundsSetZero': <OutOfBoundsPolicy.OutOfBoundsSetZero: 1>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class OutputDerivMappedStokes_1(OutputStokes_1):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def deriv_map(self) -> dict[str, numpy.ndarray[numpy.float64[m, n]]]: ...
    @property
    def radiance(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @property
    def surface_deriv_map(self) -> dict[str, numpy.ndarray[numpy.float64[m, n]]]: ...

class OutputDerivMappedStokes_3(OutputStokes_3):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def deriv_map(self) -> dict[str, numpy.ndarray[numpy.float64[m, n]]]: ...
    @property
    def radiance(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @property
    def surface_deriv_map(self) -> dict[str, numpy.ndarray[numpy.float64[m, n]]]: ...

class OutputIdealStokes_1(OutputStokes_1):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def d_radiance(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def radiance(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...

class OutputIdealStokes_3(OutputStokes_3):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def d_radiance(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def radiance(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...

class OutputStokes_1:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...

class OutputStokes_3:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...

class SingleScatterSource:
    """
    Members:

      NoSource

      Exact

      Table

      DiscreteOrdinates
    """

    DiscreteOrdinates: typing.ClassVar[
        SingleScatterSource
    ]  # value = <SingleScatterSource.DiscreteOrdinates: 2>
    Exact: typing.ClassVar[
        SingleScatterSource
    ]  # value = <SingleScatterSource.Exact: 0>
    NoSource: typing.ClassVar[
        SingleScatterSource
    ]  # value = <SingleScatterSource.NoSource: 3>
    Table: typing.ClassVar[
        SingleScatterSource
    ]  # value = <SingleScatterSource.Table: 1>
    __members__: typing.ClassVar[
        dict[str, SingleScatterSource]
    ]  # value = {'NoSource': <SingleScatterSource.NoSource: 3>, 'Exact': <SingleScatterSource.Exact: 0>, 'Table': <SingleScatterSource.Table: 1>, 'DiscreteOrdinates': <SingleScatterSource.DiscreteOrdinates: 2>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SnowKokhanovskyStokes_1(BRDFStokes_1):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class SnowKokhanovskyStokes_3(BRDFStokes_3):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    @property
    def num_deriv(self) -> int:
        """
        Number of derivatives this BRDF will calculate.
        """

class StokesBasis:
    """
    Members:

      Standard

      Solar

      Observer
    """

    Observer: typing.ClassVar[StokesBasis]  # value = <StokesBasis.Observer: 2>
    Solar: typing.ClassVar[StokesBasis]  # value = <StokesBasis.Solar: 1>
    Standard: typing.ClassVar[StokesBasis]  # value = <StokesBasis.Standard: 0>
    __members__: typing.ClassVar[
        dict[str, StokesBasis]
    ]  # value = {'Standard': <StokesBasis.Standard: 0>, 'Solar': <StokesBasis.Solar: 1>, 'Observer': <StokesBasis.Observer: 2>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SurfaceDerivativeMapping:
    interp_dim: str
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, arg0: int, arg1: int) -> None: ...
    def set_zero(self) -> None: ...
    @property
    def d_brdf(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def d_emission(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def interpolator(self) -> numpy.ndarray[numpy.float64[m, n]] | None: ...
    @interpolator.setter
    def interpolator(self, arg1: numpy.ndarray[numpy.float64[m, n]]) -> None: ...

class SurfaceStokes_1:
    brdf: BRDFStokes_1
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def get_derivative_mapping(self, name: str) -> SurfaceDerivativeMapping: ...
    @property
    def albedo(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def brdf_args(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def d_brdf_args(self) -> list[numpy.ndarray[numpy.float64[m, n]]]: ...
    @property
    def derivative_mappings(self) -> dict[str, SurfaceDerivativeMapping]: ...
    @property
    def emission(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @property
    def max_azimuthal_order(self) -> int: ...

class SurfaceStokes_3:
    brdf: BRDFStokes_3
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def get_derivative_mapping(self, name: str) -> SurfaceDerivativeMapping: ...
    @property
    def albedo(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def brdf_args(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    @property
    def d_brdf_args(self) -> list[numpy.ndarray[numpy.float64[m, n]]]: ...
    @property
    def derivative_mappings(self) -> dict[str, SurfaceDerivativeMapping]: ...
    @property
    def emission(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @property
    def max_azimuthal_order(self) -> int: ...

class TangentAltitudeSolar(ViewingGeometryBase):
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(
        self,
        tangent_altitude_m: float,
        relative_azimuth: float,
        observer_altitude_m: float,
        cos_sza: float,
    ) -> None:
        """
        Defines a viewing ray from the observer altitude, and tangent point parameters. Note that all of
        these parameters assume straight line paths (i.e. no atmospheric refraction)

        Parameters
        ----------
        tangent_altitude_m: float
            Tangent altitude in [m]
        relative_azimuth: float
            Relative azimuth angle to the sun. An angle of 0 degrees corresponds to the forward scattering plane. [rad]
        observer_altitude_m: float
            Observer altitude relative to the earth [m]
        cos_sza: float
            Cosine of the solar zenith angle at the tangent point [unitless]
        """

class ThreadingModel:
    """
    Members:

      Wavelength

      Source
    """

    Source: typing.ClassVar[ThreadingModel]  # value = <ThreadingModel.Source: 1>
    Wavelength: typing.ClassVar[
        ThreadingModel
    ]  # value = <ThreadingModel.Wavelength: 0>
    __members__: typing.ClassVar[
        dict[str, ThreadingModel]
    ]  # value = {'Wavelength': <ThreadingModel.Wavelength: 0>, 'Source': <ThreadingModel.Source: 1>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class ViewingGeometry:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self) -> None: ...
    def add_ray(self, arg0: ViewingGeometryBase) -> None: ...
    @property
    def observer_rays(self) -> list: ...

class ViewingGeometryBase:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __repr__(self) -> str: ...

class WignerD:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs): ...
    def __init__(self, m: int, n: int) -> None:
        """
        Performs calculations of the Wigner (small) d function, :math:`d^l_{m, n}(\theta)`.

        First, this class is constructed for a given `m` and `n`, and then :py:meth:`d` is called
        for a given `l`.

        The Wigner functions are closely related to the associated Legendre polynomials,
        .. math::

            d^l_{m, 0}(\theta) = \sqrt{\frac{(l-m)!}{(l+m)!}} P^m_l(\cos \theta)

        and the regular Legendre polynomials,
        .. math::

            d^l_{0, 0}(\theta) = P_l(\cos \theta)

        Parameters
        ----------
        m: int
            The parameter `m` in :math:`d^l_{m, n}`

        n: int
            The parameter `n` in :math:`d^l_{m, n}`
        """

    def d(
        self, theta: numpy.ndarray[numpy.float64], l: numpy.ndarray[numpy.int32]
    ) -> typing.Any:
        """
        Calculates :math:`d^l_{m, n}(\theta)` for a given `l`, and `m`, `n` provided in the constructor.
        Note that only one of `theta`, `l` can be array-like, one must be scalar.

        Parameters
        ----------
        theta: numpy.ndarray[numpy.float64]
            Angles (in radians) to calculate the function at

        l: numpy.ndarray[numpy.int32]
            The parameter `n` in :math:`d^l_{m, n}`

        Returns
        -------
        np.array
            The calculated Wigner function, either scalar or the same size as `theta` or `l`, whichever is array-like.
        """

def voigt_broaden(
    line_center: numpy.ndarray[numpy.float64[m, 1]],
    line_intensity: numpy.ndarray[numpy.float64[m, 1]],
    lower_energy: numpy.ndarray[numpy.float64[m, 1]],
    gamma_air: numpy.ndarray[numpy.float64[m, 1]],
    gamma_self: numpy.ndarray[numpy.float64[m, 1]],
    delta_air: numpy.ndarray[numpy.float64[m, 1]],
    n_air: numpy.ndarray[numpy.float64[m, 1]],
    iso_id: numpy.ndarray[numpy.int32[m, 1]],
    partitions: numpy.ndarray[numpy.float64[m, n], numpy.ndarray.flags.f_contiguous],
    molecular_mass: numpy.ndarray[numpy.float64[m, 1]],
    pressure: numpy.ndarray[numpy.float64[m, 1]],
    pself: numpy.ndarray[numpy.float64[m, 1]],
    temperature: numpy.ndarray[numpy.float64[m, 1]],
    wavenumber_grid: numpy.ndarray[numpy.float64[m, 1]],
    result: numpy.ndarray[
        numpy.float64[m, n],
        numpy.ndarray.flags.writeable,
        numpy.ndarray.flags.f_contiguous,
    ],
    line_contribution_width: float = 10.0,
    cull_factor: float = 0.0,
    num_threads: int = 1,
) -> None: ...

ConstantSpacing: GridSpacing  # value = <GridSpacing.ConstantSpacing: 0>
Disabled: InputValidationMode  # value = <InputValidationMode.Disabled: 2>
DiscreteOrdinates: (
    SingleScatterSource  # value = <SingleScatterSource.DiscreteOrdinates: 2>
)
Ellipsoidal: GeometryType  # value = <GeometryType.Ellipsoidal: 3>
Exact: SingleScatterSource  # value = <SingleScatterSource.Exact: 0>
LinearInterpolation: (
    InterpolationMethod  # value = <InterpolationMethod.LinearInterpolation: 1>
)
LinearSpacing: GridSpacing  # value = <GridSpacing.LinearSpacing: 1>
LowerInterpolation: (
    InterpolationMethod  # value = <InterpolationMethod.LowerInterpolation: 2>
)
NoSource: EmissionSource  # value = <EmissionSource.NoSource: 1>
Observer: StokesBasis  # value = <StokesBasis.Observer: 2>
OutOfBoundsExtend: OutOfBoundsPolicy  # value = <OutOfBoundsPolicy.OutOfBoundsExtend: 0>
OutOfBoundsSetZero: (
    OutOfBoundsPolicy  # value = <OutOfBoundsPolicy.OutOfBoundsSetZero: 1>
)
PlaneParallel: GeometryType  # value = <GeometryType.PlaneParallel: 0>
PseudoSpherical: GeometryType  # value = <GeometryType.PseudoSpherical: 1>
ShellInterpolation: (
    InterpolationMethod  # value = <InterpolationMethod.ShellInterpolation: 0>
)
Solar: StokesBasis  # value = <StokesBasis.Solar: 1>
Source: ThreadingModel  # value = <ThreadingModel.Source: 1>
Spherical: GeometryType  # value = <GeometryType.Spherical: 2>
Standard: InputValidationMode  # value = <InputValidationMode.Standard: 1>
Strict: InputValidationMode  # value = <InputValidationMode.Strict: 0>
SuccessiveOrders: (
    MultipleScatterSource  # value = <MultipleScatterSource.SuccessiveOrders: 1>
)
Table: SingleScatterSource  # value = <SingleScatterSource.Table: 1>
TwoStream: MultipleScatterSource  # value = <MultipleScatterSource.TwoStream: 2>
Wavelength: ThreadingModel  # value = <ThreadingModel.Wavelength: 0>



from __future__ import annotations

import logging
import urllib.request
import zipfile
from pathlib import Path

import xarray as xr

from .base import CachedDatabase


class WebDatabase(CachedDatabase):
    def __init__(
        self, url: str, db_root: Path | None = None, rel_path: Path | None = None
    ) -> None:
        """
        A web database is a database that is downloaded from a URL and stored.  It must
        consist of a single file.

        Parameters
        ----------
        url : str
            URL Location for the file
        db_root : Path | None, optional
            Database root directory. Optional, by default None which will set it to the config file database_root
        rel_path : Path | None, optional
            Relative path to place the file inside the db_root, by default None which will place it in the root.
        """
        super().__init__(db_root)
        self._url = url

        # Get the directory to place the data files
        if rel_path is not None:
            self._data_directory = self._db_root.joinpath(rel_path)
        else:
            self._data_directory = self._db_root

        self._data_directory.mkdir(parents=True, exist_ok=True)

        # Get the filename from the url
        self._filename = Path(url).name

    def _post_process(self):
        """
        Called after the file is downloaded. Used by derived classes to post process the file
        """

    def load(self, **kwargs):
        """
        Downloads the database from the URL if it doesn't exist
        """
        output_file = self._data_directory.joinpath(self._filename)

        if output_file.exists():
            # Already have the database downloaded
            return

        try:
            urllib.request.urlretrieve(
                self._url,
                filename=output_file.as_posix(),
            )
        except Exception as e:
            logging.exception(e)

        self._post_process()

    def clear(self):
        output_file = self._data_directory.joinpath(self._filename)

        if output_file.exists():
            output_file.unlink()

    def output_file(self):
        return self._data_directory.joinpath(self._filename)

    def path(self, key: str, **kwargs) -> Path | None:
        return self._data_directory.joinpath(key)

    def load_ds(self, key: str, **kwargs):
        return xr.open_dataset(self.path(key, **kwargs))


class ZipWebDatabase(WebDatabase):
    def __init__(
        self, url: str, db_root: Path | None = None, rel_path: Path | None = None
    ) -> None:
        """
        A web database that is a zip file. It is downloaded from a URL and then extracted.

        Parameters
        ----------
        url : str
            url of the zip file
        db_root : Path | None, optional
            Database root directory. Optional, by default None which will set it to the config file database_root
        rel_path : Path | None, optional
            Relative path to place the file inside the db_root, by default None which will place it in the root.
        """
        super().__init__(url, db_root, rel_path)

    def _post_process(self):
        output_file = self._data_directory.joinpath(self._filename)

        try:
            with zipfile.ZipFile(output_file.as_posix(), "r") as zip_ref:
                zip_ref.extractall(self._db_root)
        except Exception as e:
            logging.exception(e)


class StandardDatabase(CachedDatabase):
    def __init__(self, version: str = "latest") -> None:
        """
        Standard databases that are downloaded from the sasktran website.

        Parameters
        ----------
        version : str, optional
            Specific version to use, by default "latest"
        """
        self._url = f"https://arg.usask.ca/sasktranfiles/sasktran2_db/v_{version}/"

    def path(self, key: str, **kwargs) -> Path | None:
        db = WebDatabase(self._url + key, rel_path=Path(key).parent)
        db.load()

        return db.output_file()

    def load_ds(self, key: str, **kwargs):
        return xr.open_dataset(self.path(key, **kwargs))

    def clear(self):
        msg = "The standard database can only be cleared manually"
        raise NotImplementedError(msg)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_wf_basic():
    """
    Runs a full calculation across multiple lines of sight and wavelengths, we don't verify the results, just check the chain
    """
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    engine = sk.Engine(config, geometry, viewing_geo)

    wavel = np.arange(280.0, 800.0, 1)

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
    )

    engine.calculate_radiance(atmosphere)



from __future__ import annotations

import numpy as np

from sasktran2.atmosphere import Atmosphere

from ..base import Constituent
from . import LambertianStokes_1, LambertianStokes_3, WavelengthInterpolatorMixin


class LambertianSurface(Constituent, WavelengthInterpolatorMixin):
    def __init__(
        self,
        albedo: np.array,
        wavelengths_nm: np.array = None,
        wavenumbers_cminv: np.array = None,
        out_of_bounds_mode="zero",
    ) -> None:
        """
        A Lambertian surface that is defined by albedo at discrete grid points.

        This can either operate in a "scalar" mode where the albedo is constant in wavelength,
        a "native" mode where the albedo is defined on the same grid as the atmosphere, or
        an "interpolated" mode where the albedo is interpolated either in wavenumber or wavelength

        Parameters
        ----------
        albedo : np.array
            Surface albedo.  Can be a scalar to indicate it is constant in wavelength.  If set to an
            array it must either match the atmosphere wavelength grid dimension, or one of
            wavelengths_nm or wavenumbers_cminv must be set.
        wavelengths_nm : np.array, optional
            Wavelengths in [nm] that the albedo is specified at, by default None
        wavenumbers_cminv : np.array, optional
            Wavenumbers in inverse cm that the albedo is specified at, by default None
        out_of_bounds_mode : str, optional
            One of ["extend" or "zero"], "extend" will extend the last/first value if we are
            interpolating outside the grid. "zero" will set the albedo to 0 outside of the
            grid boundaries, by default "zero"
        """
        Constituent.__init__(self)
        WavelengthInterpolatorMixin.__init__(
            self,
            wavelengths_nm=wavelengths_nm,
            wavenumbers_cminv=wavenumbers_cminv,
            out_of_bounds_mode=out_of_bounds_mode,
            param_length=len(np.atleast_1d(albedo)),
        )
        self._albedo = np.atleast_1d(albedo)

    @property
    def albedo(self) -> np.array:
        return self._albedo

    @albedo.setter
    def albedo(self, albedo: np.array):
        self._albedo = np.atleast_1d(albedo)

    def add_to_atmosphere(self, atmo: Atmosphere):
        atmo.surface.brdf = (
            LambertianStokes_1() if atmo.nstokes == 1 else LambertianStokes_3()
        )

        interp_matrix = self._interpolating_matrix(atmo)

        atmo.surface.brdf_args[0, :] = interp_matrix @ self._albedo
        atmo.surface.d_brdf_args[0][0, :] = 1

    def register_derivative(self, atmo: Atmosphere, name: str):
        # Start by constructing the interpolation matrix
        interp_matrix = self._interpolating_matrix(atmo)

        derivs = {}

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_albedo")
        deriv_mapping.d_brdf[:] = np.ones((atmo.num_wavel, 1))
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        return derivs



# ruff: noqa: F401
from __future__ import annotations

from .amf import AirMassFactor
from .brdf.kokhanovsky import SnowKokhanovsky
from .brdf.lambertiansurface import LambertianSurface
from .brdf.modis import MODIS
from .collisioninducedabsorber import CollisionInducedAbsorber
from .emission import SurfaceThermalEmission, ThermalEmission
from .numdenscatterer import ExtinctionScatterer, NumberDensityScatterer
from .rayleigh import Rayleigh
from .solar import SolarIrradiance
from .vmraltitudeabsorber import VMRAltitudeAbsorber



from __future__ import annotations

import abc
from dataclasses import dataclass

import numpy as np

from sasktran2.atmosphere import Atmosphere


@dataclass
class OpticalQuantities:
    extinction: np.ndarray = None
    ssa: np.ndarray = None
    a1: np.ndarray = None
    a2: np.ndarray = None
    a3: np.ndarray = None
    a4: np.ndarray = None
    b1: np.ndarray = None
    b2: np.ndarray = None


class OpticalProperty(abc.ABC):
    @abc.abstractmethod
    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        pass

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:  # noqa: ARG002
        return {}

    def cross_sections(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> OpticalQuantities:
        msg = "Not Supported"
        raise NotImplementedError(msg)

    def cross_section_derivatives(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> dict:
        return {}



from __future__ import annotations

import numpy as np
import pandas as pd

from sasktran2._core import (
    Geodetic,
    GroundViewingSolar,
    TangentAltitudeSolar,
    ViewingGeometryBase,
)
from sasktran2.geodetic import WGS84
from sasktran2.solar import SolarGeometryHandlerBase, SolarGeometryHandlerForced


def ecef_to_sasktran2_ray(
    observer: np.array,
    look_vector: np.array,
    time: pd.Timestamp,
    geoid: Geodetic | None = None,
    solar_handler: SolarGeometryHandlerBase | None = None,
    ground_elevation: float = 0.0,
) -> ViewingGeometryBase:
    """
    Converts an observer, look vector in ECEF coordinates and time to a sasktran2 viewing ray object

    Parameters
    ----------
    observer : np.array
        Observer position in ECEF (ITRF) coordinates
    look_vector : np.array
        Local lok vector in ECEF (ITRF) coordinates
    time: pd.Timestamp
        Time of the observation.  Used to lookup the sun position if applicable.
    geoid : sk.Geodetic
        Geodetic object to use for the internal calculations. Defaults to WGS84
    solar_handler : sk.solar.SolarGeometryHandlerBase
        Object used to lookup the solar angles for each ray.  Defaults to None in which case a SolarHandlerForced object is used with 0 solar zenith and 0 solar azimuth

    Returns
    -------
    sk.ViewingGeometryBase
    """
    if solar_handler is None:
        solar_handler = SolarGeometryHandlerForced(0, 0)

    if geoid is None:
        geoid = WGS84()

    # Start by determining the observer's geodetic coordinates
    geoid.from_xyz(observer)

    # obs_lat = geoid.latitude
    # obs_lon = geoid.longitude
    obs_alt = geoid.altitude

    # Now start by checking if this is a limb viewing ray
    geoid.from_tangent_point(observer, look_vector)

    if geoid.altitude > ground_elevation:
        # Limb viewing
        solar_zenith, solar_azimuth = solar_handler.target_solar_angles(
            geoid.latitude, geoid.longitude, geoid.altitude, time
        )

        # Get the viewing azimuth angle
        viewing_azimuth = -np.rad2deg(
            np.arctan2(
                np.dot(look_vector, geoid.local_west),
                -np.dot(look_vector, geoid.local_south),
            )
        )

        return TangentAltitudeSolar(
            geoid.altitude,
            np.deg2rad(solar_azimuth - viewing_azimuth),
            obs_alt,
            np.cos(np.deg2rad(solar_zenith)),
        )
    # ground viewing
    geoid.from_xyz(
        geoid.altitude_intercepts(ground_elevation, observer, look_vector)[0]
    )
    solar_zenith, solar_azimuth = solar_handler.target_solar_angles(
        geoid.latitude, geoid.longitude, geoid.altitude, time
    )

    cos_viewing_zenith = np.dot(look_vector, geoid.local_up)

    if np.abs(cos_viewing_zenith) > (1 - 1e-8):
        # Basically nadir viewing and so we can't get an observer azimuth
        viewing_azimuth = 0.0
    else:
        # Get the viewing azimuth angle
        viewing_azimuth = -np.arctan2(
            np.dot(look_vector, geoid.local_west),
            -np.dot(look_vector, geoid.local_south),
        )

    return GroundViewingSolar(
        np.cos(np.deg2rad(solar_zenith)),
        np.deg2rad(solar_azimuth - viewing_azimuth),
        cos_viewing_zenith,
        obs_alt,
    )



from __future__ import annotations

import numpy as np

import sasktran2 as sk
from sasktran2.optical.rayleigh import rayleigh_cross_section_bates
from sasktran2.polarization import LegendreStorageView

from .base import Constituent


class Rayleigh(Constituent):
    def __init__(self, method: str = "bates", method_kwargs: dict | None = None):
        """
        An implementation of Rayleigh scattering.  Cross sections (and depolarization factors) can be
        calculated multiple ways, with the default method being that of 'bates'.

        Rayleigh scattering number density is estimated through the ideal gas law.

        This Constituent requires that the atmosphere object have `temperature_k`, `pressure_pa`, and
        `wavelength_nm` are all defined inside the :py:class:`sasktran2.Atmosphere` object.

        Parameters
        ----------
        method : str, optional
            Method to use to calculate the cross section.  Supported methods are
            ['bates'], by default 'bates'

        method_kwargs: dict, optional
            kwargs that can be passed to the method

        Raises
        ------
        ValueError
            If input method is not supported
        """
        self._method_kwargs = method_kwargs
        if method.lower() == "bates":
            self._rayleigh_cross_fn = rayleigh_cross_section_bates
            if self._method_kwargs is not None:
                self._fn_kwargs = self._method_kwargs
            else:
                self._fn_kwargs = {}
        elif method.lower() == "manual":

            def temp_fn(wv_micron: np.array):
                xs = np.interp(
                    wv_micron * 1000,
                    self._method_kwargs["wavelength_nm"],
                    self._method_kwargs["xs"],
                )
                king = np.interp(
                    wv_micron * 1000,
                    self._method_kwargs["wavelength_nm"],
                    self._method_kwargs["king_factor"],
                )

                return xs, king

            self._rayleigh_cross_fn = temp_fn
            self._fn_kwargs = {}
        else:
            msg = "Method must be bates or manual"
            raise ValueError(msg)

        self._ray_ext = None
        self._delta = None

    def add_to_atmosphere(self, atmo: sk.Atmosphere):
        """
        Parameters
        ----------
        atmo : sk.Atmosphere


        :meta private:
        """
        if atmo.wavelengths_nm is None:
            msg = "It is required to give the Atmosphere object wavelengths to use the Rayleigh constituent"
            raise ValueError(msg)

        if atmo.pressure_pa is None:
            msg = "It is required to set the pressure_pa property in the Atmosphere object to use the Rayleigh Constituent"
            raise ValueError(msg)

        if atmo.temperature_k is None:
            msg = "It is required to set the temperature_k property in the Atmosphere object to use the Rayleigh Constituent"
            raise ValueError(msg)

        # Get the number density from the atmosphere object at the grid points
        num_dens = atmo.state_equation.air_numberdensity["N"]

        scattering_xs, king_factor = self._rayleigh_cross_fn(
            atmo.wavelengths_nm / 1000, **self._fn_kwargs
        )

        # King factor F = (6 + 3 d) / (6 - 7 d)
        # So then (6 - 7 d) * F = 6 + 3 d
        # d (3 + 7 F) = 6(F - 1)
        # d = 6 (F - 1) / (3 + 7 F)
        self._delta = 6 * (king_factor - 1) / (3 + 7 * king_factor)

        self._ray_ext = np.outer(num_dens, scattering_xs)
        # Start by adding in the extinction
        atmo.storage.total_extinction += self._ray_ext

        # SSA temporarily stores the scattering extinction
        atmo.storage.ssa += self._ray_ext

        atmo.leg_coeff.a1[0] += self._ray_ext
        atmo.leg_coeff.a1[2] += (
            self._ray_ext * ((1 - self._delta) / (2 + self._delta))[np.newaxis, :]
        )

        if atmo.nstokes == 3:
            atmo.leg_coeff.a2[2] += (
                self._ray_ext
                * 6
                * ((1 - self._delta) / (2 + self._delta))[np.newaxis, :]
            )

            atmo.leg_coeff.b1[2] += (
                self._ray_ext
                * np.sqrt(6.0)
                * ((1 - self._delta) / (2 + self._delta))[np.newaxis, :]
            )
        elif atmo.nstokes == 4:
            msg = (
                "NSTOKES={} not currently implemented for Rayleigh constituent".format(
                    atmo.nstokes
                )
            )
            raise ValueError(msg)

    def register_derivative(self, atmo: sk.Atmosphere, name: str):
        num_dens = atmo.state_equation.air_numberdensity
        N = num_dens["N"]

        derivs = {}

        # dI/dP = dI/dN * dN/dP, and dI/dN is the extinction
        # It's easier to just treat this as a number density derivative using the Interpolated Derivative Mapping
        # and an Identity matrix as the mapping matrix

        xs = self._ray_ext / N[:, np.newaxis]

        deriv_names = []
        d_vals = []
        if atmo.calculate_pressure_derivative:
            deriv_names.append("pressure_pa")
            d_vals.append(num_dens["dN_dP"])
        if atmo.calculate_temperature_derivative:
            deriv_names.append("temperature_k")
            d_vals.append(num_dens["dN_dT"])

        for deriv_name, vert_factor in zip(deriv_names, d_vals, strict=True):
            mapping = atmo.storage.get_derivative_mapping(f"wf_{name}_{deriv_name}")
            mapping.d_extinction[:] += xs
            mapping.d_ssa[:] += (
                xs * (1 - atmo.storage.ssa) / atmo.storage.total_extinction
            )
            mapping.d_leg_coeff[:] -= atmo.storage.leg_coeff
            mapping.scat_factor[:] += xs / (
                atmo.storage.ssa * atmo.storage.total_extinction
            )

            mapping.interpolator = np.eye(len(N)) * vert_factor[np.newaxis, :]
            mapping.interp_dim = "altitude"
            mapping.assign_name = f"wf_{deriv_name}"

            d_leg_coeff = LegendreStorageView(mapping.d_leg_coeff, atmo.nstokes)

            d_leg_coeff.a1[0][:] += 1
            d_leg_coeff.a1[2][:] += ((1 - self._delta) / (2 + self._delta))[
                np.newaxis, :
            ]

            if atmo.nstokes >= 3:
                d_leg_coeff.a2[2][:] += (
                    6 * ((1 - self._delta) / (2 + self._delta))[np.newaxis, :]
                )

                d_leg_coeff.b1[2][:] += (
                    np.sqrt(6.0)
                    * ((1 - self._delta) / (2 + self._delta))[np.newaxis, :]
                )

        return derivs



from __future__ import annotations

import abc

import numpy as np
import xarray as xr
from scipy import sparse

import sasktran2 as sk


def map_derivative(mapping, np_deriv: np.ndarray, dims: list[str]) -> xr.DataArray:
    if mapping.interpolator is None:
        return xr.DataArray(np_deriv, dims=dims)
    return (
        xr.DataArray(
            (
                np_deriv.reshape(-1, np_deriv.shape[-1])
                @ sparse.csc_matrix(mapping.interpolator)
            ).reshape(np.concatenate((np_deriv.shape[:-1], [-1]))),
            dims=dims,
        )
        .transpose(*[dims[-1], *dims[:-1]])
        .rename({dims[-1]: mapping.interp_dim})
    )


def map_surface_derivative(
    mapping, np_deriv: np.ndarray, dims: list[str]
) -> xr.DataArray:
    if mapping.interpolator is None:
        return xr.DataArray(np_deriv, dims=dims)
    return xr.DataArray(
        np.einsum(
            "ijk, il->lijk",
            np_deriv,
            mapping.interpolator,
            optimize=True,
        ),
        dims=[mapping.interp_dim, *dims],
    )


class Output(abc.ABC):
    """
    An abstract class which defines how output from the radiative transfer Engine is handled.

    The Engine first calls :py:meth:`~internal_output` to get a `pybind11` object which can be passed
    to the internal radiative transfer model. After the calculation is completed, :py:meth:`~post_process`
    is called to get the exact output container that is passed back to the user.

    """

    @abc.abstractmethod
    def internal_output(self):
        pass

    @abc.abstractmethod
    def post_process(
        self,
        atmo: sk.Atmosphere,
        geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        pass


class OutputIdeal(Output):
    def __init__(self, nstokes: int):
        """
        The default output container used by SASKTRAN2.  Here radiances (and derivatives) are returned
        back densely for every input line of sight and wavelength without modification.

        Parameters
        ----------
        nstokes : int
            Number of stokes parameters to store for the output quantities

        Raises
        ------
        ValueError
            If nstokes is not 1 or 3
        """
        self._nstokes = nstokes

        if nstokes == 1:
            self._output = sk.OutputIdealStokes_1()
        elif nstokes == 3:
            self._output = sk.OutputIdealStokes_3()
        else:
            msg = "nstokes must be 1 or 3"
            raise ValueError(msg)

    def internal_output(self) -> sk.OutputIdealStokes_1 | sk.OutputIdealStokes_3:
        """
        The internal output object that can be passed to the internal engine

        Returns
        -------
        Union[sk.OutputIdealStokes_1, sk.OutputIdealStokes_3]
        """
        return self._output

    def post_process(
        self,
        atmo: sk.Atmosphere,
        geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        """
        Converts the raw output values to a more usable format.  Also performs the mapping of raw model
        derivatives to derivatives of the constituent input quantities.

        Parameters
        ----------
        atmo : sk.Atmosphere
            Atmosphere object after calculation has been performed
        geometry : sk.Geometry1D
            Geometry object
        viewing_geo : sk.ViewingGeometry
            Viewing geometry

        Returns
        -------
        xr.Dataset
        """
        # TODO: A lot of this reshaping and organizing should be done inside C++, not here
        # Maybe some of this should go into the derivative mapping classes too...
        # Have to be a little careful since this can be a slow part of the code

        # Reshape and organize output
        result = xr.Dataset()
        if atmo.wavelengths_nm is not None:
            result.coords["wavelength"] = atmo.wavelengths_nm

        result.coords["stokes"] = ["I", "Q", "U", "V"][: atmo.nstokes]
        result.coords["altitude"] = geometry.altitudes()

        radiance = self._output.radiance.reshape(
            -1, len(viewing_geo.observer_rays), atmo.nstokes
        )
        result["radiance"] = (["wavelength", "los", "stokes"], radiance)

        natmo_grid = atmo.storage.total_extinction.shape[0]
        nwavel = radiance.shape[0]

        d_radiance_raw = self._output.d_radiance.reshape(
            nwavel, len(viewing_geo.observer_rays), atmo.nstokes, -1
        )
        d_radiance_k = d_radiance_raw[:, :, :, :natmo_grid]
        d_radiance_ssa = d_radiance_raw[:, :, :, natmo_grid : 2 * natmo_grid]

        d_radiance_albedo = d_radiance_raw[:, :, :, -atmo.surface.brdf.num_deriv :]

        dk_scaled_by_dk = 1 - atmo.storage.f * atmo.unscaled_ssa
        dssa_scaled_by_dssa = (
            1 - atmo.storage.f
        ) / dk_scaled_by_dk + atmo.storage.f * atmo.storage.ssa / dk_scaled_by_dk

        for deriv_name, mapping in atmo.storage.derivative_mappings.items():
            if mapping.assign_name != "":
                name_to_place_result = mapping.assign_name
            else:
                name_to_place_result = deriv_name

            np_deriv = (
                mapping.d_extinction * dk_scaled_by_dk
                - atmo.storage.f * mapping.d_ssa * atmo.unscaled_extinction
            ).T[:, np.newaxis, np.newaxis, :] * d_radiance_k + (
                mapping.d_ssa * dssa_scaled_by_dssa
            ).T[
                :, np.newaxis, np.newaxis, :
            ] * d_radiance_ssa

            if mapping.is_scattering_derivative:
                d_radiance_scat = d_radiance_raw[
                    :,
                    :,
                    :,
                    (2 + mapping.scat_deriv_index)
                    * natmo_grid : (3 + mapping.scat_deriv_index)
                    * natmo_grid,
                ]

                np_deriv += (
                    d_radiance_scat
                    * mapping.scat_factor.T[:, np.newaxis, np.newaxis, :]
                )

                if atmo.applied_delta_m_order:
                    # Change in f for the scatterer
                    d_f = (
                        atmo.storage.d_f[:, :, mapping.scat_deriv_index]
                        * mapping.scat_factor
                    )

                    # Change in k* due to a change in f
                    np_deriv -= (d_f * atmo.unscaled_ssa * atmo.unscaled_extinction).T[
                        :, np.newaxis, np.newaxis, :
                    ] * d_radiance_k

                    # Change in ssa* due to a change in f
                    np_deriv += (
                        d_f
                        * atmo.unscaled_ssa
                        / (1 - atmo.unscaled_ssa * atmo.storage.f)
                        * (atmo.storage.ssa - 1)
                    ).T[:, np.newaxis, np.newaxis, :] * d_radiance_ssa

            if name_to_place_result in result:
                result[name_to_place_result] += map_derivative(
                    mapping, np_deriv, ["wavelength", "los", "stokes", "altitude"]
                )
            else:
                result[name_to_place_result] = map_derivative(
                    mapping, np_deriv, ["wavelength", "los", "stokes", "altitude"]
                )

        for deriv_name, mapping in atmo.surface.derivative_mappings.items():
            name_to_place_result = deriv_name

            np_deriv = d_radiance_albedo * mapping.d_brdf[:, np.newaxis, np.newaxis, :]

            mapped_derivative = map_surface_derivative(
                mapping, np_deriv.sum(axis=-1), ["wavelength", "los", "stokes"]
            )
            result[name_to_place_result] = mapped_derivative

        return result


class OutputDerivMapped(Output):
    def __init__(self, nstokes: int):
        """
        The default output container used by SASKTRAN2.  Here radiances  are returned
        back densely for every input line of sight without modification, and derivatives are mapped
        based on the atmospheric derivative mappings.

        Parameters
        ----------
        nstokes : int
            Number of stokes parameters to store for the output quantities

        Raises
        ------
        ValueError
            If nstokes is not 1 or 3
        """
        self._nstokes = nstokes

        if nstokes == 1:
            self._output = sk.OutputDerivMappedStokes_1()
        elif nstokes == 3:
            self._output = sk.OutputDerivMappedStokes_3()
        else:
            msg = "nstokes must be 1 or 3"
            raise ValueError(msg)

    def internal_output(
        self,
    ) -> sk.OutputDerivMappedStokes_1 | sk.OutputDerivMappedStokes_3:
        """
        The internal output object that can be passed to the internal engine

        Returns
        -------
        Union[sk.OutputDerivMappedStokes_1, sk.OutputDerivMappedStokes_3]
        """
        return self._output

    def post_process(
        self,
        atmo: sk.Atmosphere,
        geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        """
        Converts the raw output values to a more usable format.  Also performs the mapping of raw model
        derivatives to derivatives of the constituent input quantities.

        Parameters
        ----------
        atmo : sk.Atmosphere
            Atmosphere object after calculation has been performed
        geometry : sk.Geometry1D
            Geometry object
        viewing_geo : sk.ViewingGeometry
            Viewing geometry

        Returns
        -------
        xr.Dataset
        """
        # Reshape and organize output
        result = xr.Dataset()
        if atmo.wavelengths_nm is not None:
            result.coords["wavelength"] = atmo.wavelengths_nm

        result.coords["stokes"] = ["I", "Q", "U", "V"][: atmo.nstokes]
        result.coords["altitude"] = geometry.altitudes()

        radiance = self._output.radiance.reshape(
            -1, len(viewing_geo.observer_rays), atmo.nstokes
        )
        result["radiance"] = (["wavelength", "los", "stokes"], radiance)

        for deriv_name, mapping in atmo.storage.derivative_mappings.items():
            if mapping.assign_name != "":
                name_to_place_result = mapping.assign_name
            else:
                name_to_place_result = deriv_name

            # Make sure to copy before reshaping because the internal array is
            # Fortran ordered
            np_deriv = (
                self._output.deriv_map[deriv_name]
                .copy()
                .reshape(
                    -1, len(viewing_geo.observer_rays), atmo.nstokes, mapping.num_output
                )
                .transpose([3, 0, 1, 2])
            )

            if name_to_place_result in result:
                result[name_to_place_result] += xr.DataArray(
                    np_deriv.copy(),
                    dims=[
                        mapping.interp_dim if mapping.interp_dim != "" else "altitude",
                        "wavelength",
                        "los",
                        "stokes",
                    ],
                )
            else:
                result[name_to_place_result] = xr.DataArray(
                    np_deriv.copy(),
                    dims=[
                        mapping.interp_dim if mapping.interp_dim != "" else "altitude",
                        "wavelength",
                        "los",
                        "stokes",
                    ],
                )
        # And the surface derivatives
        for deriv_name, mapping in atmo.surface.derivative_mappings.items():
            name_to_place_result = deriv_name

            # Make sure to copy before reshaping because the internal array is
            # Fortran ordered
            np_deriv = (
                self._output.surface_deriv_map[deriv_name]
                .copy()
                .reshape(-1, len(viewing_geo.observer_rays), atmo.nstokes)
            )

            mapped_derivative = map_surface_derivative(
                mapping, np_deriv, ["wavelength", "los", "stokes"]
            )
            if mapping.interp_dim == "dummy":
                mapped_derivative = mapped_derivative.isel(**{mapping.interp_dim: 0})
            result[name_to_place_result] = mapped_derivative

        return result



from __future__ import annotations

import abc
from pathlib import Path

import xarray as xr

from sasktran2 import appconfig


class AbstractDatabase(abc.ABC):
    """
    Defines the interface for database classes within sasktran2
    """

    @abc.abstractmethod
    def path(self, key: str, **kwargs) -> Path | None:
        pass

    @abc.abstractmethod
    def load_ds(self, key: str, **kwargs) -> xr.Dataset:
        pass


class CachedDatabase(AbstractDatabase):
    def __init__(
        self, db_root: Path | None = None, rel_path: Path | None = None
    ) -> None:
        """
        A CachedDatabase is a database that caches data on the local file system.
        The root directly can optionally be specified, otherwise the default
        folder is used.

        Parameters
        ----------
        db_root : Path, optional
            The root directory to store the database, by default None
        """
        if db_root is None:
            self._db_root = appconfig.database_root()
        else:
            self._db_root = db_root

        if rel_path is not None:
            self._db_root = self._db_root.joinpath(rel_path)

        if not self._db_root.exists():
            self._db_root.mkdir(parents=True)

    @abc.abstractmethod
    def clear(self):
        """
        Deletes the entire database
        """



import numpy as np
import sasktran as sk
from sasktran.legendre import compute_greek_coefficients_legendre
from sasktran.mie.distribution import integrated_mie, sp_lognormal
from sasktran.mie.refractive import refractive_index_fn_h2so4, refractive_index_fn_water
import matplotlib.pyplot as plt
import xarray as xr
from tqdm import tqdm
from pathlib import Path


OUT_DIR = Path(
    "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/dist/cross_sections/mie/"
)


def create_table(name: str, distribution, refractive_index_fn, wavelengths):
    mie = sk.MieWiscombe()

    vals = integrated_mie(
        mie,
        distribution,
        refractive_index_fn,
        wavelengths,
        num_quad=1000,
        maxintquantile=0.99999,
    )

    # vals xs is in units of nm^2, convert to cm^2 then to m^2
    vals["xs_total"] *= 1e-14 * 1e-4
    vals["xs_scattering"] *= 1e-14 * 1e-4
    vals["xs_absorption"] *= 1e-14 * 1e-4

    lm_a1 = np.zeros_like(vals["lm_p11"].values)
    lm_a2 = np.zeros_like(lm_a1)
    lm_a3 = np.zeros_like(lm_a1)
    lm_a4 = np.zeros_like(lm_a1)
    lm_b1 = np.zeros_like(lm_a1)
    lm_b2 = np.zeros_like(lm_a1)

    for idx in range(len(vals.wavelength.values)):
        selected = vals.isel(wavelength=idx)

        (
            lm_a1[idx, :],
            lm_a2[idx, :],
            lm_a3[idx, :],
            lm_a4[idx, :],
            lm_b1[idx, :],
            lm_b2[idx, :],
        ) = compute_greek_coefficients_legendre(
            selected["lm_p11"].values,
            selected["lm_p12"].values,
            selected["lm_p11"].values,
            selected["lm_p33"].values,
            selected["lm_p34"].values,
            selected["lm_p33"].values,
        )

    vals["lm_a1"] = (["wavelength", "legendre"], lm_a1)
    vals["lm_a2"] = (["wavelength", "legendre"], lm_a2)
    vals["lm_a3"] = (["wavelength", "legendre"], lm_a3)
    vals["lm_a4"] = (["wavelength", "legendre"], lm_a4)
    vals["lm_b1"] = (["wavelength", "legendre"], lm_b1)
    vals["lm_b2"] = (["wavelength", "legendre"], lm_b2)

    ret_ds = xr.Dataset()

    ret_ds["xs_scattering"] = vals["xs_scattering"]
    ret_ds["xs_total"] = vals["xs_total"]
    ret_ds["lm_a1"] = vals["lm_a1"]
    ret_ds["lm_a2"] = vals["lm_a2"]
    ret_ds["lm_a3"] = vals["lm_a3"]
    ret_ds["lm_a4"] = vals["lm_a4"]
    ret_ds["lm_b1"] = vals["lm_b1"]
    ret_ds["lm_b2"] = vals["lm_b2"]

    return ret_ds


database_wavelength = np.arange(250, 801, 50.0)
radii = np.arange(10, 400, 10.0)

vals = []
for r in tqdm(radii):
    dist = sp_lognormal(r, 1.6)
    refrac = refractive_index_fn_h2so4()

    vals.append(create_table("sulfate_fine", dist, refrac, database_wavelength))

ds = xr.concat(vals, dim="lognormal_median_radius")
ds["lognormal_median_radius"] = radii

ds = ds.rename_dims({"wavelength": "wavelength_nm"})
ds = ds.rename_vars({"wavelength": "wavelength_nm"})
ds.to_netcdf(OUT_DIR.joinpath("sulfate_test.nc"))

pass
# dist = sp_lognormal(120, 1.3)
# refrac = refractive_index_fn_h2so4()

# create_table("sulfate_coarse", dist, refrac, database_wavelength)

# SULPHATE - STRATOSPHERIC AEROSOL
# dist = sp_lognormal(80, 1.6)
# refrac = refractive_index_fn_h2so4()

# create_table("sulfate_strat", dist, refrac, database_wavelength)

# WATER CLOUD
# dist = sp_lognormal(8000 / np.exp(5 / 2 * np.log(1.2) ** 2), 1.2)
# refrac = refractive_index_fn_water()
# create_table("water_cloud", dist, refrac, database_wavelength)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 4
    config.num_stokes = 3

    altitude_grid = np.arange(0, 65001, 5000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_wf_polarized_vmr_const():
    """
    Checks that the polarized WFs are correct for a VMR constituent
    """

    scens = _test_scenarios()

    for scen in scens:
        altitude_grid = scen["atmosphere"].model_geometry.altitudes()
        atmosphere = scen["atmosphere"]

        atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
            sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["ozone"].vmr, 0.0001, engine, atmosphere, "wf_ozone_vmr"
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_ozone_vmr"],
            radiance["wf_ozone_vmr_numeric"],
            wf_dim="ozone_altitude",
            decimal=5,
        )


def _test_scat_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    # config.single_scatter_source = sk.SingleScatterSource.NoSource
    config.num_streams = 4
    config.delta_m_scaling = False
    config.num_stokes = 3

    altitude_grid = np.arange(0, 65001, 5000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0.1, 600000, 0.6))

    wavel = np.array([600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere.surface.albedo[:] = 0.3

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_scatterer_extinction_wf_native_grid_polarized():
    """
    Tests the derivatives for a scatterer on the native grid for polarized calculation
    """

    scens = _test_scat_scenarios()

    for scen in scens:
        altitude_grid = scen["atmosphere"].model_geometry.altitudes()
        atmosphere = scen["atmosphere"]

        atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
            sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
        )

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            altitude_grid, True
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].extinction_per_m,
            1e-3,
            engine,
            atmosphere,
            "wf_strat_aerosol_extinction",
        )

        # The precision on this one can be pretty bad, but it looks right to me? Unsure
        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_extinction"],
            radiance["wf_strat_aerosol_extinction_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=2,
        )



from __future__ import annotations

import numpy as np
import xarray as xr

import sasktran2 as sk


def validate_wf(analytic, numerical, wf_dim="altitude", decimal=6):
    max_by_alt = np.abs(analytic).max(dim=wf_dim)

    max_by_alt.to_numpy()[max_by_alt.to_numpy() == 0] = 1e99

    rel_diff = (analytic - numerical) / max_by_alt

    nonzero_analytic = np.abs(analytic.to_numpy()) > 1e-99
    nonzero_numerical = np.abs(numerical.to_numpy()) > 1e-99

    np.testing.assert_array_almost_equal(
        rel_diff.to_numpy()[nonzero_analytic & nonzero_numerical], 0, decimal=decimal
    )


def numeric_wf(
    input_var: np.array,
    fractional_change: float,
    engine: sk.Engine,
    atmosphere: sk.Atmosphere,
    analytic_wf_name: str,
) -> xr.Dataset:
    base_radiance = engine.calculate_radiance(atmosphere)

    central_diff_wf = np.zeros_like(base_radiance[analytic_wf_name].to_numpy())

    for i in range(len(input_var)):
        dx = input_var[i] * fractional_change

        if dx == 0:
            dx = np.nanmean(input_var) * fractional_change

        input_var[i] += dx
        radiance_above = engine.calculate_radiance(atmosphere)

        if input_var[i] >= dx:
            # central diff
            input_var[i] -= 2 * dx
            radiance_below = engine.calculate_radiance(atmosphere)
            input_var[i] += dx

            central_diff_wf[i] = (
                radiance_above["radiance"] - radiance_below["radiance"]
            ) / (2 * dx)
        else:
            # forward diff
            central_diff_wf[i] = (
                radiance_above["radiance"] - base_radiance["radiance"]
            ) / (dx)

            input_var[i] -= dx

    base_radiance[analytic_wf_name + "_numeric"] = (
        base_radiance[analytic_wf_name].dims,
        central_diff_wf,
    )

    return base_radiance



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    # Spherical single scatter
    config = sk.Config()
    config.num_streams = 4
    config.delta_m_scaling = False

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    viewing_geo.add_ray(sk.GroundViewingSolar(0.6, 0.3, 0.5, 200000))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    # DO Preudo-spherical
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 4
    config.delta_m_scaling = False

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.PseudoSpherical,
    )

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    # DO spherical
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 4
    config.delta_m_scaling = False

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_kokhanovsky_constant():
    """
    Checks that we can create Kokhanovsky BRDF and use it in a radiative transfer calculation without errors
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.SnowKokhanovsky(
            L=3600000, M=5.5e-8
        )

        _ = engine.calculate_radiance(scen["atmosphere"])


def test_kokhanovsky_wf_scalar():
    """
    Checks the weighting functions for a snow surface relative to the parameters L and M
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.SnowKokhanovsky(
            L=3600000, M=5.5e-8
        )

        radiance = sk.test_util.wf.numeric_wf(
            scen["atmosphere"]["surface"].L,
            0.001,
            engine,
            scen["atmosphere"],
            "wf_surface_L",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_surface_L"],
            radiance["wf_surface_L_numeric"],
            "surface_wavelength",
            decimal=6,
        )

        radiance = sk.test_util.wf.numeric_wf(
            scen["atmosphere"]["surface"].M,
            0.001,
            engine,
            scen["atmosphere"],
            "wf_surface_M",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_surface_M"],
            radiance["wf_surface_M_numeric"],
            "surface_wavelength",
            decimal=6,
        )



from __future__ import annotations

import numpy as np
import xarray as xr

from sasktran2._core import Geometry1D


class ViewingGeometryContainer:
    def __init__(self, geometry_ds: xr.Dataset | None = None):
        self._geometry_ds = geometry_ds

    @property
    def geometry_ds(self) -> xr.Dataset:
        return self._geometry_ds

    def add_geometry_to_radiance(self, radiance: xr.Dataset) -> xr.Dataset:
        return xr.merge([self.geometry_ds, radiance])

    def recommended_earth_radius(self) -> float:
        return 6371000.0

    def recommended_cos_sza(self) -> float:
        raise NotImplementedError()

    def model_geometry(self, altitude_grid_m: np.ndarray) -> Geometry1D:
        raise NotImplementedError()



# ruff: noqa: F401
from __future__ import annotations

import contextlib
import importlib.util
import io
import logging

from .._core import WignerD
from . import interpolation

_cached_hapi = None  # Cache variable to store the result of the import


def get_hapi():
    global _cached_hapi  # Use the cached result if available  # noqa: PLW0603
    if _cached_hapi is not None:
        return _cached_hapi

    try:
        # Suppress print messages during import
        with contextlib.redirect_stdout(io.StringIO()):
            import hapi  # Attempt to import the package
        _cached_hapi = hapi  # Cache the imported module
    except ImportError:
        _cached_hapi = None  # Cache the failure
        logging.error(
            "Optional dependency 'hapi' is not installed. Install it to use this functionality."
        )
        raise  # Re-raise the exception

    return _cached_hapi



from __future__ import annotations

import numpy as np
import pytest
import sasktran2 as sk


def _default_settings():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    atmosphere.storage.ssa[:] = 0.9
    atmosphere.storage.total_extinction[:] = 0.1

    atmosphere.leg_coeff.a1[0, :, :] = 1

    return config, model_geometry, viewing_geo, atmosphere


def test_nan_extinction():
    config, model_geometry, viewing_geo, atmosphere = _default_settings()

    atmosphere.storage.total_extinction[10, 10] = np.nan

    engine = sk.Engine(config, model_geometry, viewing_geo)

    try:
        _ = engine.calculate_radiance(atmosphere)
        pytest.fail("Expected an exception")
    except Exception:
        pass


def test_negative_extinction():
    config, model_geometry, viewing_geo, atmosphere = _default_settings()

    atmosphere.storage.total_extinction[0, 0] = -0.1

    engine = sk.Engine(config, model_geometry, viewing_geo)

    try:
        _ = engine.calculate_radiance(atmosphere)
        pytest.fail("Expected an exception")
    except Exception:
        pass


def test_nan_ssa():
    config, model_geometry, viewing_geo, atmosphere = _default_settings()

    atmosphere.storage.ssa[10, 10] = np.nan

    engine = sk.Engine(config, model_geometry, viewing_geo)

    try:
        _ = engine.calculate_radiance(atmosphere)
        pytest.fail("Expected an exception")
    except Exception:
        pass


def test_negative_ssa():
    config, model_geometry, viewing_geo, atmosphere = _default_settings()

    atmosphere.storage.ssa[0, 0] = -0.1

    engine = sk.Engine(config, model_geometry, viewing_geo)

    try:
        _ = engine.calculate_radiance(atmosphere)
        pytest.fail("Expected an exception")
    except Exception:
        pass



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_los_refraction_refractive_one():
    # Tests that the model gives the same results with LOS refraction enabled if the refractive
    # index is forced to 1
    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    config = sk.Config()
    config.los_refraction = True

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine_refraction = sk.Engine(config, model_geometry, viewing_geo)

    radiance_refracted = engine_refraction.calculate_radiance(atmosphere)

    config = sk.Config()
    config.los_refraction = False

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    radiance = engine.calculate_radiance(atmosphere)

    np.testing.assert_allclose(
        radiance_refracted["radiance"].to_numpy(),
        radiance["radiance"].to_numpy(),
        rtol=1e-4,
    )


def test_multiple_scatter_refraction_refractive_one():
    # Tests that the model gives the same results with multiple scatter refraction enabled if the refractive
    # index is forced to 1
    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    config = sk.Config()
    config.multiple_scatter_refraction = True
    config.multiple_scatter_source = sk.MultipleScatterSource.SuccessiveOrders

    wavel = np.array([500.0])
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine_refraction = sk.Engine(config, model_geometry, viewing_geo)

    radiance_refracted = engine_refraction.calculate_radiance(atmosphere)

    config = sk.Config()
    config.multiple_scatter_refraction = False
    config.multiple_scatter_source = sk.MultipleScatterSource.SuccessiveOrders

    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    radiance = engine.calculate_radiance(atmosphere)

    np.testing.assert_allclose(
        radiance_refracted["radiance"].to_numpy(),
        radiance["radiance"].to_numpy(),
        rtol=1e-4,
    )


def test_solar_refraction_refractive_one_discrete_ordinates():
    # Tests that the model gives the same results with solar refraction enabled if the refractive
    # index is forced to 1 and we are using the discrete ordinates source
    csz = 0.1
    model_geometry = sk.Geometry1D(
        cos_sza=csz,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=csz,
        )
        viewing_geo.add_ray(ray)

    config = sk.Config()
    config.solar_refraction = True
    config.single_scatter_source = sk.SingleScatterSource.NoSource
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)
    # model_geometry.refractive_index = sk.optical.refraction.ciddor_index_of_refraction(
    #    atmosphere.temperature_k, atmosphere.pressure_pa, 0.0, 450, 600
    # )

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine_refraction = sk.Engine(config, model_geometry, viewing_geo)

    radiance_refracted = engine_refraction.calculate_radiance(atmosphere)

    config = sk.Config()
    config.solar_refraction = False
    config.single_scatter_source = sk.SingleScatterSource.NoSource
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    radiance = engine.calculate_radiance(atmosphere)

    np.testing.assert_allclose(
        radiance_refracted["radiance"].to_numpy(),
        radiance["radiance"].to_numpy(),
        rtol=1e-4,
    )



from __future__ import annotations

import numpy as np

from sasktran2.atmosphere import Atmosphere

from .base import Constituent


class AirMassFactor(Constituent):
    def __init__(
        self,
    ) -> None:
        """
        A dummy atmospheric constituent that does not add any terms to the atmosphere, but rather
        enables the calculation of Air Mass Factor derivatives.
        """
        super().__init__()

    def add_to_atmosphere(self, atmo: Atmosphere):
        pass

    def register_derivative(self, atmo: Atmosphere, name: str):  # noqa: ARG002
        altitudes = atmo.model_geometry.altitudes()

        alt_factors = -1 / np.gradient(altitudes)

        # Need to adjust bottom of atmosphere and top to match AMF definition
        alt_factors[0] *= 2
        alt_factors[-1] *= 2

        derivs = {}

        deriv_mapping = atmo.storage.get_derivative_mapping("air_mass_factor")
        deriv_mapping.d_extinction[:] += alt_factors[:, np.newaxis]
        deriv_mapping.d_ssa[:] += (
            alt_factors[:, np.newaxis]
            * (0 - atmo.storage.ssa)
            / atmo.storage.total_extinction
        )
        deriv_mapping.log_radiance_space = True

        return derivs



from __future__ import annotations

import numpy as np

import sasktran2 as sk
from sasktran2.atmosphere import Atmosphere
from sasktran2.constants import K_BOLTZMANN, PLANCK, SPEED_OF_LIGHT

from .base import Constituent


def planck_blackbody_radiance(
    temperature_k: np.ndarray, wavelengths_nm: np.ndarray
) -> np.ndarray:
    """
    Calculates the Planck function for a given set of temperatures and wavelengths.

    Parameters
    ----------
    temperature_k : np.ndarray
        Temperatures in [K]
    wavelengths_nm : np.ndarray
        Wavelengths in [nm] to calculate the radiance at

    Returns
    -------
    np.ndarray
        The blackbody radiance with units of W / (m^2 nm sr). The shape of the returned
        array is shape(len(temperature_k), len(wavelengths_nm)).
    """
    wavelengths_m = wavelengths_nm * 1e-9
    exponent = (
        PLANCK
        * SPEED_OF_LIGHT
        / (wavelengths_m * K_BOLTZMANN * temperature_k[:, np.newaxis])
    )

    return (
        (2 * PLANCK * SPEED_OF_LIGHT**2 / wavelengths_m**5)
        / (np.exp(exponent) - 1)
        * 1e-9
    )


def d_planck_blackbody_radiance_d_temperature(
    temperature_k: np.ndarray, wavelengths_nm: np.ndarray
) -> np.ndarray:
    """
    Calculates the derivative of the Planck function for a given set of temperatures and wavelengths
    with respect to the temperature parameter

    Parameters
    ----------
    temperature_k : np.ndarray
        Temperatures in [K]
    wavelengths_nm : np.ndarray
        Wavelengths in [nm] to calculate the radiance at

    Returns
    -------
    np.ndarray
        The derivative of the blackbody radiance with respect to temperature with units of W / (m^2 nm sr K). The shape of the returned
        array is shape(len(temperature_k), len(wavelengths_nm)).
    """
    wavelengths_m = wavelengths_nm * 1e-9
    exponent = (
        PLANCK
        * SPEED_OF_LIGHT
        / (wavelengths_m * K_BOLTZMANN * temperature_k[:, np.newaxis])
    )

    return (
        (2 * PLANCK * SPEED_OF_LIGHT**2 / wavelengths_m**5)
        / (np.exp(exponent) - 1) ** 2
        * (
            PLANCK
            * SPEED_OF_LIGHT
            / (wavelengths_m * K_BOLTZMANN * temperature_k[:, np.newaxis] ** 2)
        )
        * np.exp(exponent)
        * 1e-9
    )


class ThermalEmission(Constituent):
    def __init__(self):
        """
        An implementation of thermal emissions calculated from the Planck function. The emission is
        calculated with units of [W / (m^2 nm sr)].

        This Constituent requires that the atmosphere object have `temperature_k` and
        `wavelength_nm` defined inside the :py:class:`sasktran2.Atmosphere` object.
        """
        return

    def add_to_atmosphere(self, atmo: sk.Atmosphere):
        """
        Parameters
        ----------
        atmo : sk.Atmosphere

        :meta private:
        """
        if atmo.wavelengths_nm is None:
            msg = "It is required to give the Atmosphere object wavelengths to use the ThermalEmission constituent"
            raise ValueError(msg)

        if atmo.temperature_k is None:
            msg = "It is required to set the temperature_k property in the Atmosphere object to use the ThermalEmission Constituent"
            raise ValueError(msg)

        # calculate radiance in W / (m^2 nm sr)
        atmo.storage.emission_source += planck_blackbody_radiance(
            atmo.temperature_k, atmo.wavelengths_nm
        )

    def register_derivative(self, atmo: sk.Atmosphere, name: str):
        derivs = {}

        deriv_mapping = atmo.storage.get_derivative_mapping(f"{name}_temperature_k")
        deriv_mapping.d_ssa[:] = 0
        deriv_mapping.d_extinction[:] = 0
        deriv_mapping.d_emission[:] += d_planck_blackbody_radiance_d_temperature(
            atmo.temperature_k, atmo.wavelengths_nm
        )
        deriv_mapping.interp_dim = "altitude"
        deriv_mapping.assign_name = "wf_temperature_k"

        return derivs


class SurfaceThermalEmission(Constituent):
    def __init__(
        self,
        temperature_k: float,
        emissivity: np.ndarray,
    ) -> None:
        """
        A thermally emissive surface that is defined by a temperature and emissivity. The
        emission is calculated as the product of emissivity and the Planck function.

        This can either operate in a "scalar" mode where the emissivity is constant in wavelength,
        or a "native" mode where the emissivity is defined on the same grid as the atmosphere.

        Parameters
        ----------
        temperature_k : float
            Surface temperature.
        emissivity : np.ndarray
            Surface emissivity. Can be a scalar to indicate it is constant in wavelength. If set to an
            array it must match the atmosphere wavelength grid dimenstion.
        """
        Constituent.__init__(self)
        self._emissivity = np.atleast_1d(emissivity)
        self._temperature_k = temperature_k

    @property
    def temperature_k(self) -> float:
        return self._temperature_k

    @temperature_k.setter
    def temperature_k(self, temperature_k: float):
        self._temperature_k = temperature_k

    @property
    def emissivity(self) -> np.ndarray:
        return self._emissivity

    @emissivity.setter
    def emissivity(self, emissivity: np.ndarray):
        self._emissivity = np.atleast_1d(emissivity)

    def add_to_atmosphere(self, atmo: Atmosphere):
        atmo.surface.emission[:] += (
            self.emissivity
            * planck_blackbody_radiance(
                np.atleast_1d(self.temperature_k), atmo.wavelengths_nm
            ).flatten()
        )

    def register_derivative(self, atmo: sk.Atmosphere, name: str):
        planck = planck_blackbody_radiance(
            np.atleast_1d(self.temperature_k), atmo.wavelengths_nm
        )

        # Surface temperature derivative
        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_temperature_k")
        deriv_mapping.d_emission[:] = (
            self.emissivity
            * d_planck_blackbody_radiance_d_temperature(
                np.atleast_1d(self.temperature_k), atmo.wavelengths_nm
            ).T
        )
        deriv_mapping.interp_dim = "dummy"
        deriv_mapping.interpolator = np.ones((len(atmo.wavelengths_nm), 1))

        # Surface emissivity derivative
        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_emissivity")
        deriv_mapping.d_emission[:] = planck.flatten().reshape(-1, 1)
        deriv_mapping.interp_dim = "dummy"

        # If emissivity is scalear, we interpolate
        if len(self.emissivity) == 1:
            deriv_mapping.interpolator = np.ones((len(atmo.wavelengths_nm), 1))

        # Else don't need an interpolator



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.emission_source = sk.EmissionSource.Standard
    config.single_scatter_source = sk.SingleScatterSource.NoSource

    altitude_grid = np.arange(0, 65001, 5000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavelengths = np.arange(7370, 7380, 0.01)
    hitran_db = sk.optical.database.OpticalDatabaseGenericAbsorber(
        sk.database.StandardDatabase().path(
            "hitran/CH4/sasktran2/6f006bcb051f81fc57d1bd09315589bfe77b4348.nc"
        )
    )

    hitran_db = sk.database.HITRANDatabase(
        molecule="CH4",
        start_wavenumber=1355,
        end_wavenumber=1357,
        wavenumber_resolution=0.01,
        reduction_factor=1,
        backend="sasktran2",
        profile="voigt",
    )

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavelengths)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)
    atmosphere["ch4"] = sk.climatology.mipas.constituent("CH4", hitran_db)
    atmosphere["emission"] = sk.constituent.ThermalEmission()

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def _ground_test_scenarios():
    config = sk.Config()
    config.emission_source = sk.EmissionSource.Standard
    config.single_scatter_source = sk.SingleScatterSource.NoSource

    altitude_grid = np.arange(0, 65001, 5000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    viewing_geo.add_ray(sk.GroundViewingSolar(0.6, 0, 0.6, 200000))

    wavelengths = np.arange(7370, 7380, 0.01)
    hitran_db = sk.optical.database.OpticalDatabaseGenericAbsorber(
        sk.database.StandardDatabase().path(
            "hitran/CH4/sasktran2/6f006bcb051f81fc57d1bd09315589bfe77b4348.nc"
        )
    )

    hitran_db = sk.database.HITRANDatabase(
        molecule="CH4",
        start_wavenumber=1355,
        end_wavenumber=1357,
        wavenumber_resolution=0.01,
        reduction_factor=1,
        backend="sasktran2",
        profile="voigt",
    )

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavelengths)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)
    atmosphere["ch4"] = sk.climatology.mipas.constituent("CH4", hitran_db)
    atmosphere["emission"] = sk.constituent.ThermalEmission()
    atmosphere["surface_emission"] = sk.constituent.SurfaceThermalEmission(300, 0.9)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_wf_extinction_ssa_with_emission():
    """
    Checks that the WFs are correct for a VMR constituent When emissions are present
    """

    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["ch4"].vmr, 0.0001, engine, atmosphere, "wf_ch4_vmr"
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_ch4_vmr"],
            radiance["wf_ch4_vmr_numeric"],
            wf_dim="ch4_altitude",
            decimal=5,
        )


def test_wf_temperature_with_emission():
    """
    Checks that the WFs are correct for a VMR constituent When emissions are present
    """

    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere.temperature_k, 0.0001, engine, atmosphere, "wf_temperature_k"
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_temperature_k"],
            radiance["wf_temperature_k_numeric"],
            wf_dim="altitude",
            decimal=5,
        )


def test_wf_surface_temperature_with_emission():
    """
    Checks that the WFs are correct for surface temperature
    """

    scens = _ground_test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        dx = 0.1
        base_rad = engine.calculate_radiance(atmosphere)

        atmosphere["surface_emission"].temperature_k = (
            atmosphere["surface_emission"].temperature_k + dx
        )

        radiance_above = engine.calculate_radiance(atmosphere)

        atmosphere["surface_emission"].temperature_k = (
            atmosphere["surface_emission"].temperature_k - 2 * dx
        )
        radiance_below = engine.calculate_radiance(atmosphere)

        wf = (radiance_above["radiance"] - radiance_below["radiance"]) / (2 * dx)

        np.testing.assert_array_almost_equal(
            base_rad["wf_surface_emission_temperature_k"].to_numpy(), wf, decimal=5
        )


def test_wf_emissivity_with_emission():
    """
    Checks that the WFs are correct for surface emissivity
    """

    scens = _ground_test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        dx = 0.001
        base_rad = engine.calculate_radiance(atmosphere)

        atmosphere["surface_emission"].emissivity = (
            atmosphere["surface_emission"].emissivity + dx
        )

        radiance_above = engine.calculate_radiance(atmosphere)

        atmosphere["surface_emission"].emissivity = (
            atmosphere["surface_emission"].emissivity - 2 * dx
        )
        radiance_below = engine.calculate_radiance(atmosphere)

        wf = (radiance_above["radiance"] - radiance_below["radiance"]) / (2 * dx)

        np.testing.assert_array_almost_equal(
            base_rad["wf_surface_emission_emissivity"].to_numpy(), wf, decimal=5
        )



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_solar_constituent_scaling_limb():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    rad1 = engine.calculate_radiance(atmosphere)

    atmosphere["solar_irradiance"] = sk.constituent.SolarIrradiance()

    rad2 = engine.calculate_radiance(atmosphere)

    solar_irrad = sk.solar.SolarModel().irradiance(wavel)

    test = (rad2["radiance"] / rad1["radiance"]).mean(dim="los").isel(stokes=0)

    np.testing.assert_allclose(test, solar_irrad)


def test_solar_constituent_scaling_nadir():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    viewing_geo.add_ray(sk.GroundViewingSolar(0.6, 0, 0.8, 200000))

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    atmosphere["surface"] = sk.constituent.LambertianSurface(0.3)

    engine = sk.Engine(config, model_geometry, viewing_geo)

    rad1 = engine.calculate_radiance(atmosphere)

    atmosphere["solar_irradiance"] = sk.constituent.SolarIrradiance()

    rad2 = engine.calculate_radiance(atmosphere)

    solar_irrad = sk.solar.SolarModel().irradiance(wavel)

    test = (rad2["radiance"] / rad1["radiance"]).mean(dim="los").isel(stokes=0)

    np.testing.assert_allclose(test, solar_irrad)


def test_solar_constituent_scaling_DO():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.single_scatter_source = sk.SingleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.PlaneParallel,
    )

    viewing_geo = sk.ViewingGeometry()

    viewing_geo.add_ray(sk.GroundViewingSolar(0.6, 0, 0.8, 200000))

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    atmosphere["surface"] = sk.constituent.LambertianSurface(0.3)

    engine = sk.Engine(config, model_geometry, viewing_geo)

    rad1 = engine.calculate_radiance(atmosphere)

    atmosphere["solar_irradiance"] = sk.constituent.SolarIrradiance()

    rad2 = engine.calculate_radiance(atmosphere)

    solar_irrad = sk.solar.SolarModel().irradiance(wavel)

    test = (rad2["radiance"] / rad1["radiance"]).mean(dim="los").isel(stokes=0)

    np.testing.assert_allclose(test, solar_irrad)



from __future__ import annotations

import numpy as np
from scipy import interpolate
from scipy.special import roots_legendre

from sasktran2 import util


def compute_greek_coefficients(
    p11: np.ndarray,
    p12: np.ndarray,
    p22: np.ndarray,
    p33: np.ndarray,
    p34: np.ndarray,
    p44: np.ndarray,
    angle_grid: np.array,
    num_coeff: int,
):
    """
    Calculates the greek coefficients a1, a2, a3, a4, b1, b2 given the Legendre expansion of the phase function
    elements.

    Parameters
    ----------
    p11 : np.array
        Phase function P11
    p12 : np.array
        Phase function P12
    p22 : np.array
        Phase function P22
    p33 : np.array
        Phase function P33
    p34 : np.array
        Phase function P34
    p44 : np.array
        Phase function P44
    angle_grid : np.array
        Angular grid the phase functions are specified on.  Should fully span 0 to 180
    num_coeff: int
        Maximum number of coefficients to return in the expansion.

    Returns
    -------
    lm_a1 : np.array
        Greek coefficients for a1
    lm_a2 : np.array
        Greek coefficients for a2
    lm_a3 : np.array
        Greek coefficients for a3
    lm_a4 : np.array
        Greek coefficients for a4
    lm_b1 : np.array
        Greek coefficients for b1
    lm_b2 : np.array
        Greek coefficients for b2
    """
    cos_theta = np.cos(np.deg2rad(angle_grid))[::-1]

    wigner00 = util.WignerD(0, 0)
    wigner22 = util.WignerD(2, 2)
    wigner2m2 = util.WignerD(2, -2)
    wigner02 = util.WignerD(0, 2)

    wavelength_dim = p11.shape[0]
    lm_a1 = np.zeros((wavelength_dim, num_coeff))
    lm_a2 = np.zeros((wavelength_dim, num_coeff))
    lm_a3 = np.zeros((wavelength_dim, num_coeff))
    lm_a4 = np.zeros((wavelength_dim, num_coeff))
    lm_b1 = np.zeros((wavelength_dim, num_coeff))
    lm_b2 = np.zeros((wavelength_dim, num_coeff))

    c = 0.995
    nodes, weights = roots_legendre(num_coeff)

    nodes_left = (c - (-1)) / 2 * nodes + (c + (-1)) / 2
    weights_left = (c - (-1)) / 2 * weights

    nodes_right = (1 - c) / 2 * nodes + (1 + c) / 2
    weights_right = (1 - c) / 2 * weights
    all_weights = np.concatenate([weights_left, weights_right])
    cos_angle_grid = np.concatenate([nodes_left, nodes_right])

    lpoly_00 = np.zeros((num_coeff, 2 * num_coeff))
    lpoly_22 = np.zeros((num_coeff, 2 * num_coeff))
    lpoly_2m2 = np.zeros((num_coeff, 2 * num_coeff))
    lpoly_02 = np.zeros((num_coeff, 2 * num_coeff))

    for i in range(num_coeff):
        lpoly_00[i] = wigner00.d(np.arccos(cos_angle_grid), i)
        lpoly_00[i] *= all_weights / (2.0 / (2.0 * i + 1))
        lpoly_22[i] = wigner22.d(np.arccos(cos_angle_grid), i)
        lpoly_22[i] *= all_weights / (2.0 / (2.0 * i + 1))
        lpoly_2m2[i] = wigner2m2.d(np.arccos(cos_angle_grid), i)
        lpoly_2m2[i] *= all_weights / (2.0 / (2.0 * i + 1))
        lpoly_02[i] = wigner02.d(np.arccos(cos_angle_grid), i)
        lpoly_02[i] *= all_weights / (2.0 / (2.0 * i + 1))

    p11_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p11, axis=1))
    )(cos_angle_grid)
    p12_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p12, axis=1))
    )(cos_angle_grid)
    p22_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p22, axis=1))
    )(cos_angle_grid)
    p33_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p33, axis=1))
    )(cos_angle_grid)
    p34_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p34, axis=1))
    )(cos_angle_grid)
    p44_interp = interpolate.PchipInterpolator(
        cos_theta, np.transpose(np.flip(p44, axis=1))
    )(cos_angle_grid)

    lm_a1 = lpoly_00 @ p11_interp
    lm_a4 = lpoly_00 @ p44_interp

    lm_b1 = lpoly_02 @ p12_interp
    lm_b2 = lpoly_02 @ p34_interp * -1

    temp_1 = lpoly_22 @ (p22_interp + p33_interp)
    temp_2 = lpoly_2m2 @ (p22_interp - p33_interp)
    lm_a2 = (temp_1 + temp_2) / 2
    lm_a3 = (temp_1 - temp_2) / 2

    return (
        np.transpose(lm_a1),
        np.transpose(lm_a2),
        np.transpose(lm_a3),
        np.transpose(lm_a4),
        np.transpose(lm_b1),
        np.transpose(lm_b2),
    )
